from __future__ import annotations
import os
import logging

import pint
from pint import UnitRegistry, Quantity

cache_folder = os.path.join(os.path.dirname(__file__), "pint-cache")
ureg = UnitRegistry(cache_folder=cache_folder)
ureg.define("m3 = m**3")
ureg.define("m2 = m**2")
ureg.define("dm2 = dm**2")
ureg.define("cm2 = cm**2")
ureg.define("LMH = mm/h")
ureg.define("wt = 1")
ureg.define("vol = 1")
ureg.define("AU = [absorbance]")
Q_ = Quantity

logger = logging.getLogger(__name__)

# https://en.wikipedia.org/wiki/International_System_of_Units
# defines all the units we accept, each with their quantity name (similar to pint dimensionality).
QUANTITY_UNIT_MAP = {
    'time': ['s', 'min', 'h', 'ms'],                # SI quantities
    'length': ['m', 'cm'],
    'area': ['m**2', 'm2', 'dm2', 'cm2'],
    'mass': ['kg', 'g'],
    'density': ['kg/L', 'g/L'],
    'temperature': ['degC', 'degF', 'K', '°C', '°F'],
    'amount_of_substance': ['mol'],
    'volume': ['L', 'mL'],                          # Derived quantities
    'flow': ['L/h', 'L/min', 'L/d'],
    'frequency': ['Hz', 'kHz'],
    'pressure': ['Pa', 'bar', 'pascal'],            # Note: pint prefers pascal over Pa so we define both.
    'mass flow rate': ['kg/h', 'g/s', 'g/min', 'g/h'],
    'electrical conductance': ['mS/cm'],
    'percentage': ['%', 'vol%', 'wt%', 'mol%'],     # Custom quantity but supported by pint
    'column volume': ['CV'],                        # Custom quantity
    'absorbance': ['AU', 'mAU', 'milliAU'],
    'permeability': ['LMH/bar', 'L/m2/h/bar', 'L/h/m2/bar'],
    'flux': ['LMH', 'L/m2/h', 'L/h/m2'],
}
""" Map quantity names to unit names. """

QUANTITY_PINT_MAP: dict[str, str] = {
    'time': '[time]',
    'length': '[length]',
    'area': '[length] ** 2',
    'mass': '[mass]',
    'density': '[mass] / [length] ** 3',
    'flow': '[length] ** 3 / [time]',
    'volume': '[length] ** 3',
    'frequency': '1 / [time]',
    'temperature': '[temperature]',
    'amount_of_substance': '[substance]',
    'pressure': '[mass] / [length] / [time] ** 2',
    'mass flow rate': '[mass] / [time]',
    'electrical conductance': '[current] ** 2 * [time] ** 3 / [mass] / [length] ** 3',
    'percentage': '[percentage]',
    'flux': '[length] / [time]',
    'permeability': '[length] ** 2 * [time] / [mass]',
    'absorbance': '[absorbance]',
}
""" Map quantity names to pint dimensions or None if not a pint dimension. """

FLOAT_COMPARE_DELTA = 0.01
""" The precision to use when comparing float values for equality. """

# build list of all supported units
_supported_units: list[None | str] = [None]
for v in QUANTITY_UNIT_MAP.values():
    _supported_units.extend(v)

# These units are not stored as a tag unit but as a tag value in the Base tag,
# so they need special treatment.
# Note that this list is a total static list. The actual valid units depend on the
# totalizers that are defined in the uod.
BASE_VALID_UNITS = ['L', 'h', 'min', 's', 'mL', 'CV', 'DV', 'g', 'kg']


def get_supported_units() -> list[str | None]:
    return list(_supported_units)

def is_supported_unit(unit: str) -> bool:
    return unit in _supported_units

def get_volume_units():
    return QUANTITY_UNIT_MAP['volume']

def convert_value_to_unit(value: float | int, source_unit: str, target_unit: str) -> float:
    """ Convert a value with a unit to the value in another unit.

    Raises ValueError if the units are not compatible.
    """
    try:
        val = ureg.Quantity(value, source_unit).to(target_unit).magnitude
        return val
    except pint.DimensionalityError:
        raise ValueError(f"Cannot convert between units '{source_unit}' and '{target_unit}'")

def get_unit_quantity_name(unit: str) -> str:
    """ For a supported unit that is not None, return its quantity name. Raises ValueError if the unit is not supported. """
    for key in QUANTITY_UNIT_MAP.keys():
        if unit in QUANTITY_UNIT_MAP[key]:
            return key
    raise ValueError(f"Invalid unit: '{unit}'")

def _get_quantity_name_for_pint_dim(pint_dimension: str) -> str | None:
    for key in QUANTITY_PINT_MAP.keys():
        vals = QUANTITY_PINT_MAP[key]
        if vals is not None:
            if pint_dimension == vals:
                return key

def get_compatible_unit_names(unit: str | None) -> list[str]:
    """ For a supported unit, return its compatible units. If the unit is not supported, a ValueError is raised. """
    if unit is None:
        return [""]

    quantity_name = get_unit_quantity_name(unit)
    assert quantity_name is not None and quantity_name != ""

    pint_mapping = QUANTITY_PINT_MAP.get(quantity_name)
    if pint_mapping is None:
        # non-pint unit, return all units of this quantity name
        return QUANTITY_UNIT_MAP[quantity_name]

    pu = _get_pint_unit(unit)
    if pu is None:
        return [""]
    elif unit == "%":
        return ['%', 'vol%', 'wt%', 'mol%']
    elif pu.dimensionless:
        return [""]
    else:
        dims = pu.dimensionality
        quantity_name = _get_quantity_name_for_pint_dim(str(dims))
        if quantity_name is not None:
            return QUANTITY_UNIT_MAP[quantity_name]
        raise NotImplementedError(f"Pint unit '{pu}' with dimensionality '{str(dims)}' has no defined quantity name")

def are_comparable(unit_a: str | None, unit_b: str | None) -> bool:
    """ Determine whether two supported units are comparable, i.e. have the same quantity name.

    For example, 'cm' and 'm' are comparable because they have the same quantity name, 'length'.

    If one or both units are not supported, raise ValueError.
    """
    if unit_a == unit_b:
        return True
    if unit_a is None or unit_b is None:
        return False
    if get_unit_quantity_name(unit_a) != get_unit_quantity_name(unit_b):
        return False
    compatibles_a = get_compatible_unit_names(unit_a)
    result = unit_b in compatibles_a
    return result

def compare_values(op: str, value_a: str, unit_a: str | None, value_b: str, unit_b: str | None) -> bool:
    """ Compare two values using the provided operator (<, <=, >, >=, =, ==, !=) and return the comparison result.

    If units are given, the units are checked for comparability. The units may be different as long as they are comparable.

    In case of error, a ValueError is raised.
    """
    if not are_comparable(unit_a, unit_b):
        raise ValueError(f"Cannot compare values with incompatible units '{unit_a}' and '{unit_b}'")

    def is_pint():
        if unit_a is not None and unit_a != unit_b:
            quantity_name = get_unit_quantity_name(unit_a)
            assert quantity_name is not None and quantity_name != ""

            pint_mapping = QUANTITY_PINT_MAP.get(quantity_name, None)
            if pint_mapping is None:
                # different non-pint units cannot currently be compared. To support these, we must provide a conversion
                # to the base (SI) unit that can be applied before comparison. We don't currently need this.
                raise NotImplementedError(f"Custom comparison of non-pint units {unit_a} and {unit_b} is not supported")
            else:
                return True
        return False

    is_pint_units = is_pint()
    if is_pint_units:
        fval_a, fval_b = as_float(value_a), as_float(value_b)
        if fval_a is None:
            raise ValueError("Cannot compare values, first value is missing or not numeric")
        if fval_b is None:
            raise ValueError("Cannot compare values, second value is missing or not numeric")
        quantity_a = value_a if unit_a is None else ureg.Quantity(fval_a, unit_a)
        quantity_b = value_b if unit_b is None else ureg.Quantity(fval_b, unit_b)
    else:
        assert unit_a == unit_b, f"Units should be the same but are not, '{unit_a}' vs '{unit_b}'"
        quantity_a = value_a
        quantity_b = value_b

    # we can now ignore units and conversion - either they're None or equal - or they are
    # pint units for which Quantity handles comparison automatically

    # infer when to do numeric vs string comparison. this is only relevant for non-pint units
    if not is_pint_units:
        if op in ['<', '<=', '>', '>=']:  # these operators only make sence for numeric values
            assert isinstance(quantity_a, str), f"quantity_a '{quantity_a}' should be a str here"
            assert isinstance(quantity_b, str), f"quantity_b '{quantity_b}' should be a str here"
            quantity_a, quantity_b = as_float(quantity_a), as_float(quantity_b)
            if quantity_a is None:
                raise ValueError("Cannot compare values, first value is missing or not numeric")
            if quantity_b is None:
                raise ValueError("Cannot compare values, second value is missing or not numeric")
        else:
            # if both are numerical, perform a numerical comparison
            assert isinstance(quantity_a, str), f"quantity_a '{quantity_a}' should be a str here"
            assert isinstance(quantity_b, str), f"quantity_b '{quantity_b}' should be a str here"
            quantity_a_f, quantity_b_f = as_float(quantity_a), as_float(quantity_b)
            if quantity_a_f is not None and quantity_b_f is not None:
                quantity_a, quantity_b = quantity_a_f, quantity_b_f

    try:
        match op:
            case '<':
                result = quantity_a < quantity_b  # type: ignore
            case '<=':
                result = quantity_a <= quantity_b  # type: ignore
            case '=' | '==':
                if isinstance(quantity_a, Quantity) and isinstance(quantity_b, Quantity):
                    diff = quantity_a - quantity_b
                    result = abs(diff.magnitude) < FLOAT_COMPARE_DELTA
                elif isinstance(quantity_a, float) and isinstance(quantity_b, float):
                    result = abs(quantity_a - quantity_b) < FLOAT_COMPARE_DELTA
                else:
                    result = quantity_a == quantity_b
            case '>':
                result = quantity_a > quantity_b  # type: ignore
            case '>=':
                result = quantity_a >= quantity_b  # type: ignore
            case '!=':
                result = quantity_a != quantity_b
            case _:
                raise ValueError(f"Invalid operator: '{op}'")
    except TypeError:
        msg = "Conversion type error for values {!r} and {!r}".format(value_a, value_b)
        logger.error(msg, exc_info=True)
        raise ValueError("Conversion error")

    assert isinstance(result, bool), f"Comparison result was not type bool but {type(result)}"
    return result

def as_float(value: str) -> float | None:
    """ Parse string value as a float and return it. If the value is not a number, return None. """
    try:
        return float(value)
    except Exception:
        return None

def as_int(value: str) -> int | None:
    """ Parse string value as int and return it. If the value is not an int, return None. """
    fval = as_float(value)
    if fval is not None:
        if fval.is_integer():
            return int(fval)
    return None

def _get_pint_unit(unit: str | None) -> pint.Unit | None:
    if unit is None:
        return None
    try:
        return ureg.Unit(unit)
    except Exception:
        logger.error(f"Could not get pint unit for unit '{unit}'")
        raise



import logging

import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
import openpectus.aggregator.models as Mdl
from openpectus.aggregator.aggregator import Aggregator

logger = logging.getLogger(__name__)


class AggregatorMessageHandlers:
    def __init__(self, aggregator: Aggregator):
        self.aggregator = aggregator
        aggregator.dispatcher.set_register_handler(self.handle_RegisterEngineMsg)
        aggregator.dispatcher.set_disconnect_handler(self.handle_EngineDisconnected)
        aggregator.dispatcher.set_connect_handler(self.handle_EngineConnected)
        aggregator.dispatcher.set_message_handler(EM.UodInfoMsg, self.handle_UodInfoMsg)
        aggregator.dispatcher.set_message_handler(EM.TagsUpdatedMsg, self.handle_TagsUpdatedMsg)
        aggregator.dispatcher.set_message_handler(EM.RunLogMsg, self.handle_RunLogMsg)
        aggregator.dispatcher.set_message_handler(EM.ControlStateMsg, self.handle_ControlStateMsg)
        aggregator.dispatcher.set_message_handler(EM.MethodStateMsg, self.handle_MethodStateMsg)
        aggregator.dispatcher.set_message_handler(EM.ErrorLogMsg, self.handle_ErrorLogMsg)
        aggregator.dispatcher.set_message_handler(EM.RunStartedMsg, self.handle_RunStartedMsg)
        aggregator.dispatcher.set_message_handler(EM.RunStoppedMsg, self.handle_RunStoppedMsg)

    async def handle_RegisterEngineMsg(self, register_engine_msg: EM.RegisterEngineMsg) -> AM.RegisterEngineReplyMsg:
        """ Registers engine """
        if register_engine_msg.secret != self.aggregator.secret:
            logger.info(
                f'Engine registration message {register_engine_msg} received ' +
                "but secret does not match aggregator secret"
            )
            return AM.RegisterEngineReplyMsg(success=False, secret_match=False)
        engine_id = self.aggregator.create_engine_id(register_engine_msg)
        if self.aggregator.dispatcher.has_connected_engine_id(engine_id):
            logger.error(
                f"""Registration failed for engine_id {engine_id}. An engine with that engine_id already
                has a websocket connection. """
            )
            return AM.RegisterEngineReplyMsg(success=False, engine_id=engine_id, secret_match=True)

        # TODO consider how to handle registrations
        # - disconnect/reconnect should work
        # - client kill/reconnect should work
        # - engine_id reused with "same uod" should take over session, else fail as misconfigured client
        # - add machine name + uod secret

        # initialize client data
        if not self.aggregator.has_registered_engine_id(engine_id):
            self.aggregator.from_engine.register_engine_data(
                Mdl.EngineData(
                    engine_id=engine_id,
                    computer_name=register_engine_msg.computer_name,
                    uod_name=register_engine_msg.uod_name,
                    uod_author_name=register_engine_msg.uod_author_name,
                    uod_author_email=register_engine_msg.uod_author_email,
                    uod_filename=register_engine_msg.uod_filename,
                    location=register_engine_msg.location,
                    engine_version=register_engine_msg.engine_version,
                )
            )

        logger.info(f"Registration of engine {engine_id} successful")
        return AM.RegisterEngineReplyMsg(success=True, engine_id=engine_id, secret_match=True)

    def validate_msg(self, msg: EM.EngineMessage):
        if not self.aggregator.has_registered_engine_id(msg.engine_id):
            return AM.ErrorMessage(message=f'No engine registered under id {msg.engine_id}')
        # possibly more validations...
        return None

    async def handle_UodInfoMsg(self, msg: EM.UodInfoMsg) -> AM.SuccessMessage | AM.ErrorMessage:
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        logger.debug(f"Got UodInfo from client: {str(msg)}")
        self.aggregator.from_engine.uod_info_changed(
            msg.engine_id, msg.readings,
            msg.commands, msg.plot_configuration,
            msg.hardware_str, msg.required_roles,
            msg.data_log_interval_seconds)
        return AM.SuccessMessage()

    async def handle_EngineConnected(self, engine_id: str):
        self.aggregator.from_engine.engine_connected(engine_id)

    async def handle_EngineDisconnected(self, engine_id: str):
        self.aggregator.from_engine.engine_disconnected(engine_id)

    async def handle_TagsUpdatedMsg(self, msg: EM.TagsUpdatedMsg) -> AM.SuccessMessage | AM.ErrorMessage:
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        logger.debug(f"Got tags update from client: {str(msg)}")
        self.aggregator.from_engine.tag_values_changed(msg.engine_id, msg.tags)
        return AM.SuccessMessage()

    async def handle_RunLogMsg(self, msg: EM.RunLogMsg) -> AM.SuccessMessage | AM.ErrorMessage:
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        logger.debug(f"Got run log from client: {str(msg)}")
        self.aggregator.from_engine.runlog_changed(msg.engine_id, msg.run_id, msg.runlog)
        return AM.SuccessMessage()

    async def handle_ControlStateMsg(self, msg: EM.ControlStateMsg) -> AM.SuccessMessage | AM.ErrorMessage:
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        logger.debug(f"Got control state from client: {str(msg)}")
        self.aggregator.from_engine.control_state_changed(msg.engine_id, msg.control_state)
        return AM.SuccessMessage()

    async def handle_MethodStateMsg(self, msg: EM.MethodStateMsg):
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        logger.debug(f"Got method state from client: {str(msg)}")
        self.aggregator.from_engine.method_state_changed(msg.engine_id, msg.method_state)
        return AM.SuccessMessage()

    async def handle_ErrorLogMsg(self, msg: EM.ErrorLogMsg):
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        logger.debug(f'error log message from engine: {msg.log}')
        self.aggregator.from_engine.error_log_changed(msg.engine_id, msg.log)
        return AM.SuccessMessage()

    async def handle_RunStartedMsg(self, msg: EM.RunStartedMsg):
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        self.aggregator.from_engine.run_started(msg)
        return AM.SuccessMessage()

    async def handle_RunStoppedMsg(self, msg: EM.RunStoppedMsg):
        validation_errors = self.validate_msg(msg)
        if validation_errors is not None:
            return validation_errors

        self.aggregator.from_engine.run_stopped(msg)
        return AM.SuccessMessage()



# adapted from https://stackoverflow.com/a/68843198
import os
from typing import Tuple

from starlette.staticfiles import StaticFiles


class SinglePageApplication(StaticFiles):
    """Acts similar to the bripkens/connect-history-api-fallback
    NPM package."""

    def __init__(self, directory: str, index='index.html') -> None:
        self.index = index

        # set html=True to resolve the index even when no
        # the base path is passed in
        super().__init__(directory=directory, packages=None, html=True, check_dir=True)

    def lookup_path(self, path: str) -> Tuple[str, os.stat_result]:
        """Returns the index file when no match is found.

        Args:
            path (str): Resource path.

        Returns:
            [tuple[str, os.stat_result]]: Always returns a full path and stat result.
        """
        full_path, stat_result = super().lookup_path(path)

        # if a file cannot be found
        if stat_result is None:
            return super().lookup_path(self.index)

        return full_path, stat_result



from __future__ import annotations

import logging

import openpectus.engine.models as EM
from pydantic import BaseModel, ConfigDict

logger = logging.getLogger(__name__)

SystemStateEnum = EM.SystemStateEnum
SystemTagName = EM.SystemTagName
TagDirection = EM.TagDirection
MethodStatusEnum = EM.MethodStatusEnum
EntryDataType = EM.EntryDataType

class ProtocolModel(BaseModel):
    model_config = ConfigDict(from_attributes = True)


class ReadingCommand(ProtocolModel):
    """ Represents a entry command for a reading. """
    command_id: str
    name: str
    command: str
    choice_names: list[str]


class ReadingInfo(ProtocolModel):
    discriminator: str
    tag_name: str
    valid_value_units: list[str] | None
    entry_data_type: EntryDataType | None
    commands: list[ReadingCommand]
    command_options: dict[str, str] | None

    def has_command_id(self, command_id: str) -> bool:
        for c in self.commands:
            if c.command_id == command_id:
                return True
        return False


class CommandInfo(ProtocolModel):
    """ Represents a uod command that may or may not be a reading entry command. """
    name: str
    docstring: str | None


TagValueType = float | int | str | None


class TagValue(ProtocolModel):
    name: str = ""
    tick_time: float
    value: TagValueType = None
    value_unit: str | None
    value_formatted: str | None = None
    direction: TagDirection = TagDirection.Unspecified


class RunLogLine(ProtocolModel):
    id: str
    command_name: str
    start: float
    end: float | None
    progress: float | None  # between 0 and 1
    start_values: list[TagValue]
    end_values: list[TagValue]
    forcible: bool | None = None
    cancellable: bool | None = None
    forced: bool | None = None
    cancelled: bool | None = None


class RunLog(ProtocolModel):
    lines: list[RunLogLine]

    @staticmethod
    def empty() -> RunLog:
        return RunLog(lines=[])


class MethodLine(ProtocolModel):
    id: str
    content: str


class Method(ProtocolModel):
    lines: list[MethodLine]

    @staticmethod
    def empty() -> Method:
        return Method(lines=[])

    @staticmethod
    def from_pcode(pcode: str) -> Method:
        method = Method.empty()
        line_num: int = 1
        for line in pcode.splitlines():
            method.lines.append(MethodLine(id=f"id_{line_num}", content=line))
            line_num += 1
        return method

    def as_pcode(self) -> str:
        pcode = '\n'.join([line.content for line in self.lines])
        return pcode


class MethodState(ProtocolModel):
    started_line_ids: list[str]
    executed_line_ids: list[str]
    injected_line_ids: list[str]

    @staticmethod
    def empty() -> MethodState:
        return MethodState(started_line_ids=[], executed_line_ids=[], injected_line_ids=[])


class ControlState(ProtocolModel):
    is_running: bool
    is_holding: bool
    is_paused: bool


class PlotColorRegion(ProtocolModel):
    process_value_name: str
    # color string compatible with css e.g.: '#aa33bb', 'rgb(0,0,0)', 'rgba(0,0,0,0)', 'red'
    value_color_map: dict[str | int | float, str]


class PlotAxis(ProtocolModel):
    label: str
    process_value_names: list[str]
    y_max: int | float
    y_min: int | float
    color: str


class SubPlot(ProtocolModel):
    axes: list[PlotAxis]
    ratio: int | float


class PlotConfiguration(ProtocolModel):
    process_value_names_to_annotate: list[str]
    color_regions: list[PlotColorRegion]
    sub_plots: list[SubPlot]
    x_axis_process_value_names: list[str]

    @staticmethod
    def empty() -> PlotConfiguration:
        return PlotConfiguration(
            process_value_names_to_annotate=[],
            color_regions=[],
            sub_plots=[],
            x_axis_process_value_names=[]
        )


class ErrorLogEntry(ProtocolModel):
    message: str
    created_time: float
    severity: int


class ErrorLog(ProtocolModel):
    entries: list[ErrorLogEntry]

    @staticmethod
    def empty() -> ErrorLog:
        return ErrorLog(entries=[])



import asyncio
import logging
from argparse import ArgumentParser, BooleanOptionalAction
import importlib.util
from logging.handlers import RotatingFileHandler
from os import path
import pathlib
from typing import Literal
from itertools import chain
import sys

import multiprocess

from openpectus import log_setup_colorlog, sentry, __version__, build_number
from openpectus.engine.engine import Engine
from openpectus.engine.engine_message_handlers import EngineMessageHandlers
from openpectus.engine.engine_message_builder import EngineMessageBuilder
from openpectus.engine.hardware import NullHardware
from openpectus.engine.hardware_recovery import ErrorRecoveryConfig, ErrorRecoveryDecorator
from openpectus.lang.exec.tags import SystemTagName, TagCollection
from openpectus.lang.exec.uod import UnitOperationDefinitionBase
from openpectus.protocol.engine_dispatcher import EngineDispatcher
from openpectus.engine.engine_runner import EngineRunner

from openpectus.test.engine.utility_methods import EngineTestRunner

log_setup_colorlog()

StateKind = Literal["Started", "Connected", "Disconnected", "Reconnecting", "Reconnected"]

# Sphinx fails auto generating docs because __file__ is not defined
# when it runs the code.
if locals().get("__file__", None):
    file_log_path = path.join(pathlib.Path(__file__).parent.resolve(), 'openpectus-engine.log')
    file_handler = RotatingFileHandler(file_log_path, maxBytes=2*1024*1024, backupCount=5)
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(logging.Formatter('%(asctime)s [%(levelname)s]: %(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
    logging.root.addHandler(file_handler)

logger = logging.getLogger("openpectus.engine.engine")
logger.setLevel(logging.INFO)

logging.getLogger("openpectus.lang.exec.pinterpreter").setLevel(logging.INFO)
logging.getLogger("openpectus.protocol.engine_dispatcher").setLevel(logging.INFO)
logging.getLogger("openpectus.engine.engine_runner").setLevel(logging.INFO)
logging.getLogger("openpectus.engine.internal_commands_impl").setLevel(logging.INFO)
logging.getLogger("asyncua.client").setLevel(logging.WARNING)


default_host = "127.0.0.1"
default_port = "9800"
default_port_secure = "443"


def get_arg_parser():
    parser = ArgumentParser("Start Pectus Engine")
    parser.add_argument("-ahn", "--aggregator_hostname", required=False, default=default_host,
                        help="Aggregator websocket host name. Default is 127.0.0.1")
    parser.add_argument("-ap", "--aggregator_port", required=False, default=None,
                        help=f"Aggregator websocket port number. Default is {default_port} or {default_port_secure} " +
                        "if using --secure")
    parser.add_argument("-s", "--secure", action=BooleanOptionalAction,
                        help="Access aggregator using https/wss rather than http/ws")
    parser.add_argument("-uod", "--uod", required=False, default="openpectus/engine/configuration/demo_uod.py",
                        help="Filename of the UOD")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-validate", "--validate", action=BooleanOptionalAction,
                       help="Run Uod validation and exit. Cannot be used with -rd")
    group.add_argument("-rd", "--show_register_details", action=BooleanOptionalAction,
                       help="Show register details for UOD authoring and exit. Cannot be used with -validate")
    parser.add_argument("-sev", "--sentry_event_level", required=False,
                        default=sentry.EVENT_LEVEL_DEFAULT, choices=sentry.EVENT_LEVEL_NAMES,
                        help=f"Minimum log level to send as sentry events. Default is '{sentry.EVENT_LEVEL_DEFAULT}'")
    parser.add_argument("-secret", "--secret", required=False, default="",
                        help="Secret used to get access to aggregator")
    return parser


engine: Engine | None = None
runner: EngineRunner | None = None


def run_validations(uod: UnitOperationDefinitionBase) -> bool:
    try:
        logger.info("Verifying hardware configuration and connection")
        uod.validate_configuration()
        uod.hwl.validate_offline()
        uod.hwl.connect()
        uod.hwl.validate_online()
        logger.info("Hardware validation successful")

        logger.info("Building uod commands")
        uod.build_commands()
        return True
    except Exception:
        logger.error("An hardware related error occurred. Engine cannot start.")
        return False

async def main_async(args, loop: asyncio.AbstractEventLoop):
    global engine, runner
    try:
        uod = create_uod(args.uod)
    except Exception as ex:
        logger.error(f"Failed to create uod: {ex}")
        return

    engine = Engine(uod, enable_archiver=True)

    # if --aggregator_port is specified, use it, else select a default port based on --secure
    if args.aggregator_port is not None:
        port = args.aggregator_port
    else:
        port = default_port_secure if args.secure else default_port

    dispatcher = EngineDispatcher(f"{args.aggregator_hostname}:{port}", args.secure, uod.options, args.secret)

    if len(uod.required_roles) > 0 and not dispatcher.is_aggregator_authentication_enabled():
        logger.warning('"with_required_roles" specified in "demo_uod.py" but aggregator does ' +
                       'not support authentication. Engine will not be visible in the frontend.')

    if not run_validations(uod):
        sys.exit(1)

    sentry.set_engine_uod(uod)

    # wrap hwl with error recovery decorator
    connection_status_tag = engine._system_tags[SystemTagName.CONNECTION_STATUS]
    uod.hwl = ErrorRecoveryDecorator(uod.hwl, ErrorRecoveryConfig(), connection_status_tag)

    message_builder = EngineMessageBuilder(engine)
    # create runner that orchestrates the error recovery mechanism
    runner = EngineRunner(dispatcher, message_builder, engine.emitter, loop)
    _ = EngineMessageHandlers(engine, dispatcher)

    # TODO Possibly check dispatcher.check_aggregator_alive() and exit early


    async def on_steady_state():
        logger.info("Starting engine on first steady-state")
        assert engine is not None, "Engine was unexpectedly None on first stady_state"
        engine.run()

    runner.first_steady_state_callback = on_steady_state
    await runner.run()



async def close_async():
    global engine, runner
    logger.debug("Stopping engine components")
    if engine is not None:
        engine.stop()
    if runner is not None:
        await runner.shutdown()
    logger.debug("Engine components stopped. Exiting")


def create_uod(uod_filepath: str) -> UnitOperationDefinitionBase:
    if uod_filepath is None or uod_filepath == "" or not isinstance(uod_filepath, str):
        raise ValueError("Uod is not specified")

    try:
        spec = importlib.util.spec_from_file_location('uod', uod_filepath)
        assert spec is not None
        uod_module = importlib.util.module_from_spec(spec)
        assert spec.loader is not None
        spec.loader.exec_module(uod_module)
        logger.info(f"Imported uod from path '{uod_module.__file__}'")
    except Exception as ex:
        raise Exception(f"Failed to import uod module from path {uod_filepath}") from ex

    try:
        uod = uod_module.create()
        logger.info(f"Created uod: {uod}")
        return uod
    except Exception as ex:
        raise Exception("Failed to create uod instance") from ex


def validate_and_exit(uod_name: str):
    logger.info("Engine started in validation mode")

    try:
        uod = create_uod(uod_name)
        logger.info(f"Uod '{uod_name}' created successfully")
    except Exception:
        logger.error(f"Validation failed. Failed to create uod '{uod_name}'", exc_info=True)
        sys.exit(1)

    uod.system_tags = TagCollection.create_system_tags()

    logger.info("Validating uod configuration")
    uod.validate_configuration()

    try:
        logger.info("Running offline hardware validation")
        uod.hwl.validate_offline()
        logger.info("Offline validation successful")
    except Exception:
        logger.error("Offline validation failed", exc_info=True)
        sys.exit(1)

    run_example_commands(uod)

    logger.info("Running online hardware validation")
    try:
        logger.info("Connecting to hardware")
        uod.hwl.connect()
        logger.info("Hardware connected")
    except Exception:
        logger.info("Hardware connection failed", exc_info=True)
        sys.exit(1)

    try:
        uod.hwl.validate_online()
        logger.info("Online validation successful")
    except Exception:
        logger.error("Online validation failed", exc_info=True)

    logger.info("Validation complete. Exiting.")
    sys.exit(0)


def run_example_commands(uod: UnitOperationDefinitionBase):
    uod.build_commands()
    logger.info("Validating UOD command examples")
    uod.hwl = NullHardware()

    def run_example_with_description(description: str, example: str) -> list[str]:
        failed_cmds: list[str] = []
        try:
            runner = EngineTestRunner(uod_factory=lambda: uod, pcode=example)
            with runner.run() as instance:
                instance.start()
                # wait up to 1 minute, that ought to be enought for everybody
                instance.run_until_event("method_end", max_ticks=10*60)
                logger.debug(instance.get_runtime_table())
                logger.debug(f"{description} executed successfully")
        except Exception as ex:
            logger.error(f"{description} '{example}' failed: {str(ex)}")
            failed_cmds.append(example)
        return failed_cmds

    logger.info("Executing example commands")
    with multiprocess.Pool() as pool:  # type: ignore
        failed_cmds: list[str] = list(
          chain.from_iterable(
            pool.map(
                lambda t: run_example_with_description(t[0], t[1]),
                uod.generate_pcode_examples()
            )
          )
        )

    if len(failed_cmds) > 0:
        logger.error(f"Example commands failed: {','.join(failed_cmds)}")
    else:
        logger.info("All example commands executed succesfully")

def show_register_details_and_exit(uod_name: str):
    logger.info("Engine started in register details mode")

    try:
        uod = create_uod(uod_name)
        logger.info(f"Uod '{uod_name}' created successfully")
    except Exception:
        logger.error(f"Validation failed. Failed to create uod '{uod_name}'", exc_info=True)
        sys.exit(1)

    uod.system_tags = TagCollection.create_system_tags()

    try:
        logger.info("Connecting to hardware")
        uod.hwl.connect()
        logger.info("Hardware connected")
    except Exception:
        logger.info("Hardware connection failed", exc_info=True)
        sys.exit(1)

    try:
        uod.hwl.show_online_register_details()
    except Exception:
        logger.error("Error showing register details")
        sys.exit(1)

    sys.exit(0)


def main():
    print(f"OpenPectus Engine v. {__version__}, build: {build_number}")
    args = get_arg_parser().parse_args()
    sentry.init_engine(args.sentry_event_level)
    if args.validate:
        validate_and_exit(args.uod)
    elif args.show_register_details:
        show_register_details_and_exit(args.uod)

    logger.info("Engine starting")

    # Handling KeyboardInterrupt seems to fix the halt that sometimes otherwise occurs. But we still get the error:
    # "sys:1: RuntimeWarning: coroutine 'EngineDispatcher.disconnect_async' was never awaited"
    # https://stackoverflow.com/questions/54525836/where-do-i-catch-the-keyboardinterrupt-exception-in-this-async-setup#54528397
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(main_async(args, loop))
        logger.info("Main loop completed")
    except KeyboardInterrupt:
        logger.info("User requested engine to stop")
        loop.run_until_complete(close_async())
    except Exception:
        logger.error("Unhandled exception in main. Stopping.", exc_info=True)
        loop.run_until_complete(close_async())

    logger.info("Engine stopped")


if __name__ == "__main__":
    main()



from __future__ import annotations
from contextlib import contextmanager
import logging
import time
from typing import Callable, Generator, Literal
from openpectus.engine.models import EngineCommandEnum

from openpectus.lang.exec.clock import WallClock
from openpectus.lang.exec.runlog import RuntimeRecordStateEnum
from openpectus.lang.exec.events import BlockInfo, EventListener
from openpectus.lang.exec.timer import NullTimer
from openpectus.lang.exec.uod import UnitOperationDefinitionBase
import openpectus.protocol.models as Mdl
from openpectus.engine.engine import Engine, EngineTiming
from openpectus.lang.grammar.pgrammar import PGrammar
from openpectus.lang.model.pprogram import PProgram


logger = logging.getLogger(__name__)


def build_program(s, skip_enrich_analyzers=False) -> PProgram:
    p = PGrammar()
    p.parse(s)
    return p.build_model(skip_enrich_analyzers=skip_enrich_analyzers)


UodFactory = Callable[[], UnitOperationDefinitionBase]

RunCondition = Callable[[], bool]

EventName = Literal[
    "start", "stop", "block_start", "block_end",
    "restart", "pause", "hold",
    "method_end"
]
""" Defines the awaitable events of the test engine runner """

InstructionName = Literal[
    "Block", "End block", "End blocks",
    "Watch", "Alarm", "Mark",
    "Pause", "Hold", "Wait",
    "Stop", "Restart"
]
""" Defines the awaitable instructions of the test engine runner """

FindInstructionState = Literal[
    "any", "started", "completed", "failed", "cancelled"
]
""" Defines the awaitable instruction states of the test engine runner """


class EngineTestRunner:
    """ Expose an interface of Engine similar to what is available in the frontend to tests. """
    def __init__(self, uod_factory: UodFactory, pcode: str = "", interval: float = 0.1, speed: float = 1.0) -> None:
        self.uod_factory = uod_factory
        self.pcode = pcode
        self.interval = interval
        self.speed = speed
        logger.debug(f"Created engine test runner, speed: {self.speed:.2f}, interval: {(self.interval*1000):.2f}ms")

    @contextmanager
    def run(self) -> Generator[EngineTestInstance, None, None]:
        timing = EngineTiming(WallClock(), NullTimer(), self.interval, self.speed)
        uod = self.uod_factory()
        engine = Engine(uod, timing)
        instance = EngineTestInstance(engine, self.pcode, timing)
        try:
            yield instance
        except Exception:
            raise
        finally:
            engine.cleanup()
            del instance


class EngineTestInstance(EventListener):
    def __init__(self, engine: Engine, pcode: str, timing: EngineTiming) -> None:
        self.engine = engine
        self.timing = timing

        self.engine.run(skip_timer_start=True)
        self.set_method(pcode)

        self._search_index = 0
        self.engine.emitter.add_listener(self)  # register as listener for lifetime events, so they can be awaited
        self._last_event: EventName | None = None

    def set_method(self, pcode: str):
        method = Mdl.Method.from_pcode(pcode)
        self.engine.set_method(method)

    def start(self):
        self.engine.schedule_execution(EngineCommandEnum.START)
        self.timing.timer.start()

    @property
    def marks(self) -> list[str]:
        return self.engine.interpreter.get_marks()

    def run_until_condition(self, condition: RunCondition, max_ticks=30) -> int:
        """ Continue program until condition occurs. Return the number of ticks spent.

        Raises TimeoutError if the condition is not met before max_ticks is reached.
        """
        if condition():
            return 0

        ticks = 0
        max_ticks_scaled = max_ticks * self.timing.speed
        last_tick_time = 0.0

        while not condition():
            tick_time = time.time()

            ticks += 1
            if ticks > max_ticks_scaled:
                if max_ticks == max_ticks_scaled:
                    raise TimeoutError(f"Condition did not occur within {max_ticks} ticks")
                else:
                    raise TimeoutError(
                        f"Condition did not occur within {max_ticks_scaled} ticks, " +
                        "(scaled from {max_ticks} using speed {self.speed})"
                    )
            increment = self.timing.interval
            if last_tick_time > 0.0:
                increment = tick_time - last_tick_time
                self.engine.tick(tick_time, increment)

            last_tick_time = tick_time
            elapsed = time.time() - last_tick_time
            deadline = self.timing.interval - elapsed
            if deadline > 0.0:
                while last_tick_time+self.timing.interval-time.time() > 0:
                    time.sleep(min(last_tick_time+self.timing.interval-time.time(), 0.01))
            else:
                logger.warning(f"Sleep deadline for tick was negative: {deadline}")

            if self.engine.has_error_state():
                ex = self.engine.get_error_state_exception()
                if ex is None:
                    raise ValueError("Engine failed with an unspecified error")
                else:
                    raise ValueError(f"Engine failed with exception: {ex}")

        return ticks

    def run_until_instruction(
            self,
            instruction_name: InstructionName,
            state: FindInstructionState = "any",
            max_ticks=30,
            increment_index=True
            ) -> int:
        """ Continue program execution and wait until the given instruction is run.

        Use the `state` argument to specify how to wait. The default is 'any' which waits for any record for
        the given instruction. Note that for some instructions this happens much earlier than the start state,
        e.g. for Watch. This can be changed to e.g. `started` to wait until the instruction is started.

        Return the number of ticks spent.

        Raises TimeoutError if the instruction is not found before max_ticks is reached.

        Use `increment_index` to control searching within states of the same instruction as last search. The default
        value of True skips to the next instruction which is normally what you want. If set to False, search also
        includes the record of the previous match. This is useful if searching for different states for the same
        record, such as "started" and "completed". See also `index_step_back()`.
        """

        # convert the easy-to-use literal into one of the enum states that find_instruction needs
        request_state: RuntimeRecordStateEnum | None = None
        if state != "any":
            if state == "started":
                request_state = RuntimeRecordStateEnum.Started
            elif state == "completed":
                request_state = RuntimeRecordStateEnum.Completed
            elif state == "failed":
                request_state = RuntimeRecordStateEnum.Failed
            elif state == "cancelled":
                request_state = RuntimeRecordStateEnum.Cancelled

        def cond() -> bool:
            index = self.engine.runtimeinfo.find_instruction(instruction_name, self._search_index, request_state)
            if index is None:
                return False
            else:
                # store position so we only search from there next time
                # print(f"Found {instruction_name} at index {index}")
                self._search_index = index + 1 if increment_index else index
                if instruction_name == "Restart":  # except if restarting in which case we start over
                    self._search_index = 0
                return True

        logger.debug(f"Start waiting for instruction {instruction_name}, state: {state}")
        try:
            ticks = self.run_until_condition(cond, max_ticks=max_ticks)
        except TimeoutError:
            raise TimeoutError(f"Timeout while waiting for instruction '{instruction_name}', state: {state}")

        logger.debug(f"Done waiting for instruction {instruction_name}, state: {state}")
        return ticks

    def index_step_back(self, steps=1):
        # Not a good abstraction. Specifically not for alarms
        # the best way would be to maintain a index for each runtime record
        # os use tick as index, then search all records from there. That eliminates
        # the use of an index alltogether - but does slow it down, though, as it
        # has to search all records on every query.
        if steps < 1:
            raise ValueError("Steps must be positive")
        if self._search_index - steps < 0:
            self._search_index = 0
        else:
            self._search_index = self._search_index - steps

    def _clear_last_event(self):
        self._last_event = None

    def run_until_event(self, event_name: EventName, max_ticks=100) -> int:
        """ Continue program execution and wait until the given event fires.

        Return the number of ticks spent.

        Raises TimeoutError if the instruction is not found before max_ticks is reached.
        """
        def cond():
            if self._last_event == event_name:
                self._clear_last_event()
                return True
            else:
                return False
        try:
            return self.run_until_condition(cond, max_ticks=max_ticks)
        except TimeoutError:
            raise TimeoutError(f"Timeout while waiting for event '{event_name}'")

    def run_ticks(self, ticks: int) -> None:
        """ Continue program execution until te specified number of ticks. """

        logger.debug(f"Start waiting for {ticks} ticks")
        max_ticks = ticks + 1
        try:
            _ = self.run_until_condition(lambda: False, max_ticks)
        except TimeoutError:
            logger.debug(f"Done waiting for {ticks} ticks")
            return

        raise ValueError("Could not wait??")

    def get_runtime_table(self, description: str = "") -> str:
        """ Return a text view of the runtime table contents. """
        return self.engine.runtimeinfo.get_as_table(description)

    # --- EventListener impl ----

    def on_engine_configured(self):
        self._last_event = None

    def on_start(self, run_id: str):
        self._last_event = "start"

    def on_block_start(self, block_info: BlockInfo):
        self._last_event = "block_start"

    def on_block_end(self, block_info: BlockInfo, new_block_info: BlockInfo | None):
        self._last_event = "block_end"

    def on_tick(self, tick_time: float, increment_time: float):
        pass

    def on_method_end(self):
        self._last_event = "method_end"

    def on_stop(self):
        self._last_event = "stop"

    def on_engine_shutdown(self):
        pass

    # --- EventListener end ----


# globals for run_engine and continue_engine
last_tick_time = 0.0
interval = 0.1


def run_engine(engine: Engine, pcode: str, max_ticks: int = -1) -> int:
    global last_tick_time, interval
    print("Interpretation started")
    engine._running = True
    engine.set_method(Mdl.Method.from_pcode(pcode=pcode))
    engine.schedule_execution(EngineCommandEnum.START)

    ticks = 1
    last_tick_time = 0.0

    while True:
        tick_time = time.time()

        if ticks > max_ticks:
            print(f"Stopping because max_ticks {max_ticks} was reached")
            return ticks

        increment = 0.0
        if last_tick_time > 0.0:
            increment = tick_time - last_tick_time

        # execute tick even with 0 increment
        engine.tick(tick_time, increment)

        last_tick_time = tick_time
        elapsed = time.time() - last_tick_time
        deadline = interval - elapsed
        if deadline > 0.0:
            while last_tick_time+interval-time.time() > 0:
                time.sleep(min(last_tick_time+interval-time.time(), 0.01))
        else:
            logger.warning(f"Sleep deadline for tick was negative: {deadline}")

        if engine.has_error_state():
            ex = engine.get_error_state_exception()
            if ex is None:
                raise ValueError("Engine failed with an unspecified error")
            else:
                raise ValueError(f"Engine failed with exception: {ex}")

        ticks += 1


def continue_engine(engine: Engine, max_ticks: int = -1) -> int:
    global last_tick_time, interval
    print("Interpretation continuing")
    ticks = 1

    while True:
        tick_time = time.time()

        if ticks > max_ticks:
            print(f"Stopping because max_ticks {max_ticks} was reached")
            return ticks

        increment = 0
        if last_tick_time > 0.0:
            increment = tick_time - last_tick_time

        # execute tick even with 0 increment
        engine.tick(tick_time, increment)

        last_tick_time = tick_time
        elapsed = time.time() - last_tick_time
        deadline = interval - elapsed
        if deadline > 0.0:
            while last_tick_time+interval-time.time() > 0:
                time.sleep(min(last_tick_time+interval-time.time(), 0.01))
        else:
            logger.warning(f"Sleep deadline for tick was negative: {deadline}")

        if engine.has_error_state():
            ex = engine.get_error_state_exception()
            if ex is None:
                raise ValueError("Engine failed with an unspecified error")
            else:
                raise ValueError(f"Engine failed with exception: {ex}")

        ticks += 1


def print_runlog(e: Engine, description=""):
    runlog = e.interpreter.runtimeinfo.get_runlog()
    print(f"Runlog {runlog.id} records: ", description)
    for item in runlog.items:
        name = f"{str(item.name):<20}"
        prog = f"{item.progress:.2f}" if item.progress else ""
        print(f"{name}   {item.state:<15}    {prog}")


def print_runtime_records(e: Engine, description: str = ""):
    table = e.interpreter.runtimeinfo.get_as_table(description)
    print(table)


def configure_test_logger():
    logging.basicConfig(format=' %(name)s :: %(levelname)-8s :: %(message)s')


def set_engine_debug_logging():
    engine_modules = [
        "openpectus.engine.engine",
        # "openpectus.engine.internal_commands",
        "openpectus.engine.internal_commands_impl",
        __name__
    ]
    for m in engine_modules:
        logger = logging.getLogger(m)
        logger.setLevel(logging.DEBUG)


def set_interpreter_debug_logging():
    logger = logging.getLogger("openpectus.lang.exec.pinterpreter")
    logger.setLevel(logging.DEBUG)



import logging
import time
import unittest
from typing import Any
from openpectus.lang.exec.tags_impl import ReadingTag, SelectTag
from openpectus.engine.hardware import RegisterDirection

import pint
from openpectus.lang.exec.tags import SystemTagName, Tag, TagDirection
from openpectus.lang.exec.uod import (
    UnitOperationDefinitionBase,
    UodCommand,
    UodBuilder,
    RegexNumber,
)
from openpectus.test.engine.utility_methods import (
    EngineTestRunner,
    configure_test_logger, set_engine_debug_logging, set_interpreter_debug_logging
)


configure_test_logger()
set_engine_debug_logging()
set_interpreter_debug_logging()
logging.getLogger("openpectus.lang.exec.runlog").setLevel(logging.DEBUG)

# pint takes forever to initialize - long enough
# to throw off timing of the first instruction.
# so we initialize it first
_ = pint.Quantity("0 s")


def create_test_uod() -> UnitOperationDefinitionBase:
    def reset(cmd: UodCommand, **kvargs) -> None:
        count = cmd.get_iteration_count()
        if count == 0:
            cmd.context.tags.get("Reset").set_value("Reset", time.time())
        elif count == 4:
            cmd.context.tags.get("Reset").set_value("N/A", time.time())
            cmd.set_complete()

    def cmd_with_args(cmd: UodCommand, **kvargs) -> None:
        print("arguments: " + str(kvargs))
        cmd.set_complete()

    def cmd_arg_parse(args: str) -> dict[str, Any] | None:
        if args is None or args == "" or args == "FAIL":
            return None
        result = {}
        for i, item in enumerate(args.split(",")):
            result[f"item{i}"] = item.strip()
        return result

    def cmd_regex(cmd: UodCommand, number, number_unit) -> None:
        cmd.context.tags["CmdWithRegex_Area"].set_value(number + number_unit, time.time())
        cmd.set_complete()

    def overlap_exec(cmd: UodCommand, **kvargs) -> None:
        count = cmd.get_iteration_count()
        if count >= 9:
            cmd.set_complete()

    builder = (
        UodBuilder()
        .with_instrument("TestUod")
        .with_author("Test Author", "test@openpectus.org")
        .with_filename(__file__)
        .with_hardware_none()
        .with_location("Test location")
        .with_hardware_register("FT01", RegisterDirection.Both, path="Objects;2:System;2:FT01")
        .with_hardware_register(
            "Reset",
            RegisterDirection.Both,
            path="Objects;2:System;2:RESET",
            from_tag=lambda x: 1 if x == "Reset" else 0,
            to_tag=lambda x: "Reset" if x == 1 else "N/A",
        )
        # Readings
        .with_tag(ReadingTag("FT01", "L/h"))
        .with_tag(SelectTag("Reset", value="N/A", unit=None, choices=["Reset", "N/A"]))
        .with_tag(Tag("Danger", value=True, unit=None, direction=TagDirection.Output, safe_value=False))
        .with_command(name="Reset", exec_fn=reset)
        .with_command(name="CmdWithArgs", exec_fn=cmd_with_args, arg_parse_fn=cmd_arg_parse)
        .with_command(name="overlap1", exec_fn=overlap_exec)
        .with_command(name="overlap2", exec_fn=overlap_exec)
        .with_command_overlap(['overlap1', 'overlap2'])
        .with_tag(Tag("CmdWithRegex_Area", value=""))
        .with_command_regex_arguments(
            name="CmdWithRegexArgs",
            arg_parse_regex=RegexNumber(units=['m2']),
            exec_fn=cmd_regex)
    )
    uod = builder.build()
    uod.hwl.connect()
    return uod


class TestEngineNew(unittest.TestCase):

    def test_run_until_condition(self):
        runner = EngineTestRunner(create_test_uod)
        with runner.run() as instance:
            run_time = instance.engine.tags[SystemTagName.RUN_TIME]

            instance.start()
            start = time.time()
            instance.run_until_condition(lambda: run_time.as_float() >= 1)
            end = time.time()
            self.assertAlmostEqual(end-start, 1.1, delta=0.2)

    def test_wait(self):
        code = """
Wait: 1s
"""
        runner = EngineTestRunner(create_test_uod, code)
        with runner.run() as instance:
            instance.start()

            t = instance.run_until_instruction("Wait", "started", increment_index=False)
            self.assertEqual(4, t)
            ticks = instance.run_until_instruction("Wait", "completed")
            self.assertAlmostEqual(10, ticks, delta=2)

    def test_run_until_instruction(self):
        code = """
Mark: A
Wait: 1s
Pause: 1.5s
Mark: B
"""
        runner = EngineTestRunner(create_test_uod, code)
        with runner.run() as instance:
            instance.start()
            run_time = instance.engine.tags[SystemTagName.RUN_TIME]

            instance.run_until_instruction("Mark")
            # two ticks passed before the first "user" instruction. That is expected
            self.assertAlmostEqual(0.2, run_time.as_float(), delta=0.1)

            instance.run_until_instruction("Pause")

            # self.assertAlmostEqual(1.1, run_time.as_float(), delta=0.1)
            instance.run_until_instruction("Mark")

            with self.assertRaises(TimeoutError):
                instance.run_until_instruction("Mark")

    def test_run_until_instruction_block(self):
        code = """
Mark: A
Block: A
    Pause: 1s
    End block
    Wait: 1s
Mark: B
"""
        runner = EngineTestRunner(create_test_uod, code)
        with runner.run() as instance:
            instance.start()

            self.assertEqual([], instance.engine.interpreter.get_marks())

            instance.run_until_instruction("Block")
            self.assertEqual(['A'], instance.engine.interpreter.get_marks())

            with self.assertRaises(TimeoutError):
                instance.run_until_instruction("Wait")

            instance.run_until_instruction("Mark")

            # This is a little funky in that get_marks() uses the same record as source
            # as does run_until_instruction(). Thefore B is marked even "before" it runs
            self.assertEqual(['A', 'B'], instance.engine.interpreter.get_marks())

    def test_watch_restart(self):
        code = """
Watch: Run Counter >= 2
    Stop

Wait: 1s
Increment run counter
Restart
"""
        runner = EngineTestRunner(create_test_uod, code)

        with runner.run() as instance:
            run_counter = instance.engine.tags["Run Counter"]
            instance.start()
            # Run Counter is 0 by default, and indicates the number of completed runs
            self.assertEqual(0, run_counter.as_number())

            instance.run_until_instruction("Restart")
            self.assertEqual(1, run_counter.as_number())
            instance.run_until_event("start")

            print(instance.get_runtime_table("A"))

            instance.run_until_instruction("Restart")
            # Increment run counter should have incremented it - wtf?
            # self.assertEqual(2, run_counter.as_number())

            instance.run_until_event("start")
            instance.run_until_instruction("Watch")

            print(instance.get_runtime_table("B"))

            with self.assertRaises(TimeoutError):
                instance.run_until_instruction("Restart")

            print(instance.get_runtime_table("C"))

    def test_mark_in_alarm_body_runs_in_each_alarm_instance(self):
        code = """
Alarm: Block Time > 0s
    Mark: A
    Wait: 0.5s
"""
        runner = EngineTestRunner(create_test_uod, code)
        with runner.run() as instance:
            instance.start()
            instance.run_until_instruction("Alarm", "started")

            # print(instance.get_runtime_table("start"))

            instance.run_until_instruction("Mark", "completed")
            self.assertEqual(['A'], instance.marks)

            instance.run_ticks(9)

            # print(instance.get_runtime_table("awaiting 2nd alarm"))

            self.assertEqual(['A', 'A'], instance.marks)

    def test_watch_in_alarm_body_runs_in_each_alarm_instance(self):
        code = """
Alarm: Block Time > 0s
    Watch: Block Time > 0.5s
        Mark: A
    Wait: 0.5s
"""
        runner = EngineTestRunner(create_test_uod, code)
        with runner.run() as instance:
            instance.start()
            instance.run_until_instruction("Alarm", "started")

            print(instance.get_runtime_table("start"))

            instance.run_until_instruction("Mark", "completed")
            self.assertEqual(['A'], instance.marks)

            print(instance.get_runtime_table("mark"))

            instance.index_step_back(2)  # Wait occurs before Mark
            instance.run_until_instruction("Wait", "completed")

            instance.run_ticks(7)
            print(instance.get_runtime_table("awaiting 2nd alarm"))

            # This does not work because the alarm node is way up in the record list.
            # we would need aonther way to wait for alarm, including a new find_instruction
            # variant that locates the alarm record at indexes the states in there instead
            # of the normal one that indexes full records.
            # the same goes for any other node inside the alarm body
            # instance.run_until_instruction("Alarm", "started")

            self.assertEqual(['A', 'A'], instance.marks)


    def test_run_until_method_end(self):
        code = """
Mark: A
Mark: B
Mark: C
"""
        runner = EngineTestRunner(create_test_uod, code)
        logging.getLogger("openpectus.lang.exec.pinterpreter").setLevel(logging.DEBUG)

        with runner.run() as instance:
            instance.start()
            self.assertEqual([], instance.marks)
            instance.run_until_event("method_end")
            self.assertEqual(['A', 'B', 'C'], instance.engine.interpreter.get_marks())


    def test_run_until_stop(self):
        code = """
Mark: A
Stop
"""
        runner = EngineTestRunner(create_test_uod, code)
        logging.getLogger("openpectus.lang.exec.pinterpreter").setLevel(logging.DEBUG)

        with runner.run() as instance:
            instance.start()
            instance.run_until_event("stop")
            self.assertEqual([], instance.marks)


    @unittest.skip("speed > 1 not implemented")
    def test_speed_perceived(self):
        pcode = """
Base: s
Mark: A
Wait: 15s
15 Mark: A2
Mark: B
"""
        runner = EngineTestRunner(create_test_uod, pcode=pcode, speed=30)
        with runner.run() as instance:
            instance.start()
            run_time = instance.engine.tags[SystemTagName.RUN_TIME]
            time.sleep(1)
            self.assertAlmostEqual(run_time.as_float(), 30, delta=3)

            time.sleep(1)
            self.assertAlmostEqual(run_time.as_float(), 60, delta=3)

    @unittest.skip("speed > 1 not implemented")
    def test_speed_timer(self):
        pcode = """
Wait: 30s
Mark: B
"""
        runner = EngineTestRunner(create_test_uod, pcode=pcode, speed=30)
        with runner.run() as instance:
            run_time = instance.engine.tags[SystemTagName.RUN_TIME]
            instance.start()

            instance.run_until_instruction("Wait", "started")
            run_time.set_value(0, 0)

            instance.run_until_instruction("Mark")
            # Why the two ticks threshold?
            # Completing Wait takes a tick and 30 secs at speed 30 can vary one tick in evaluation
            run_time_value = run_time.as_float()
            self.assertAlmostEqual(33, run_time_value, delta=3)

    # speed is really important as it allows tests that wait much longer than a few ticks.
    # This allows testing timers much easier because we can use much larger periods - without
    # longer test runs. The key is that single ticks drown in long waiting times and becomes
    # irrelevant, just as they are in practice.

# --- Inject ---


    def test_inject_command(self):
        pcode = """
Mark: A
Mark: B
Mark: C
"""
        runner = EngineTestRunner(create_test_uod, pcode=pcode)
        with runner.run() as instance:
            instance.start()
            instance.run_until_instruction("Mark")
            self.assertEqual(['A'], instance.marks)

            instance.engine.inject_code("Mark: I")

            instance.run_until_condition(lambda: 'B' in instance.marks)
            self.assertEqual(['A', 'B', 'I'], instance.marks)

            instance.run_until_event("method_end")
            self.assertEqual(['A', 'B', 'I', 'C'], instance.marks)

    def test_inject_thresholds_1(self):
        pcode = """
Mark: A
0.25 Mark: B
Mark: C
"""
        runner = EngineTestRunner(create_test_uod, pcode=pcode)
        with runner.run() as instance:
            instance.start()

            instance.engine.tags[SystemTagName.BASE].set_value("s", instance.engine._tick_time)

            instance.run_until_instruction("Mark")
            self.assertEqual(['A'], instance.marks)

            instance.engine.inject_code("Mark: I")
            instance.run_until_condition(lambda: 'I' in instance.marks)
            self.assertEqual(['A', 'I'], instance.marks)

            instance.run_until_event("method_end")
            self.assertEqual(['A', 'I', 'B', 'C'], instance.marks)


    def test_inject_thresholds_2(self):
        pcode = """
Mark: A
0.2 Mark: B
Mark: C
"""
        runner = EngineTestRunner(create_test_uod, pcode=pcode)
        with runner.run() as instance:
            instance.start()

            instance.engine.tags[SystemTagName.BASE].set_value("s", instance.engine._tick_time)

            instance.run_until_instruction("Mark")
            self.assertEqual(['A'], instance.marks)

            instance.engine.inject_code("0.3 Mark: I")
            instance.run_until_condition(lambda: 'B' in instance.marks)
            self.assertEqual(['A', 'B'], instance.marks)

            instance.run_until_event("method_end")
            self.assertTrue(['A', 'B', 'C', 'I'] == instance.marks or ['A', 'B', 'I', 'C'] == instance.marks)

    def test_info_warning_error(self):
        pcode = """
Info: foo
Warning: bar
Error: baz
Stop
"""
        runner = EngineTestRunner(create_test_uod, pcode=pcode)
        with runner.run() as instance:
            instance.start()
            instance.run_until_event("stop")  # will raise on engine error



"""Change ErrorLog to AggregatedErrorLog

Revision ID: b792689d4cd4
Revises: 774e4df9ce40
Create Date: 2024-10-08 09:43:26.241195

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'b792689d4cd4'
down_revision: Union[str, None] = '774e4df9ce40'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # The only difference between ErrorLog and AggregatedErrorLog is the addition of the field
    # occurrences which has a default value of 1. It appears to work without changes.
    pass


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###




from typing import Literal
from openpectus.lang.model.pprogram import PNode


class EngineNotInitializedError(Exception):
    def __init__(self, message: str, *args: object) -> None:
        self.message = message
        super().__init__(*args)

    def __str__(self) -> str:
        return self.message or type(self).__name__


class EngineError(Exception):
    def __init__(self, message: str, user_message: str | Literal["same"] | None = None, *args) -> None:
        self.message = message
        if user_message == "same":
            self.user_message = message
        else:
            self.user_message = user_message
        super().__init__(*args)

    def __str__(self) -> str:
        return self.message or self.user_message or type(self).__name__


class UodValidationError(Exception):
    """ Raised when a UnitOperationDefinition definition/configuration error occurs. """
    pass

    def __str__(self) -> str:
        return type(self).__name__


class InterpretationError(Exception):
    """ Raised by interpreter when a general error occurs """
    def __init__(self,
                 message: str,
                 user_message: str | Literal["same"] | None = "same",
                 exception: Exception | None = None, *args: object
                 ) -> None:
        self.message: str = message
        if user_message == "same":
            self.user_message = message
        else:
            self.user_message = user_message
        self.exception = exception
        super().__init__(*args)

    def __str__(self) -> str:
        return self.message or self.user_message or type(self).__name__


class NodeInterpretationError(InterpretationError):
    """ Raised by interpreter when an instruction specific error occurs """
    def __init__(self, node: PNode, message: str, user_message: str | Literal["same"] | None = "same",
                 exception: Exception | None = None, *args: object) -> None:
        self.node = node
        self.message: str = message
        self.exception = exception
        if user_message == "same":  # avoid using base message for user messages
            user_message = message
        base_message = f"An error occurred in instruction '{node.display_name}': {message}"
        super().__init__(base_message, user_message, exception, *args)

    def __str__(self) -> str:
        return self.message or self.user_message or type(self).__name__

class InterpretationInternalError(InterpretationError):
    """ Raised by interpreter if an internal error occurs """
    def __init__(self, message: str, exception: Exception | None = None, *args: object) -> None:
        self.message: str = message
        self.exception = exception
        base_message = f"An internal error occurred. Interpretation cannot continue: {message}"
        super().__init__(base_message, "Internal error", exception, *args)

    def __str__(self) -> str:
        return self.message or self.user_message or type(self).__name__



import ctypes
from datetime import UTC, datetime
import os
import logging
import platform
import time
import csv
from typing import Callable

from openpectus.lang.exec.runlog import RunLog
from openpectus.lang.exec.tags import Tag, TagCollection

logger = logging.getLogger(__name__)

LOW_DISKSPACE_MB = 50
VERY_LOW_DISKSPACE_MB = 5

# file open defaults
encoding = 'utf-8'

# csv option defaults
delimiter = ','     # used in old system
# delimiter = ';'    # makes Excel 365 understand it out of the box
quoting = csv.QUOTE_NONE
escapechar = None
# Note:  The MarkTag value may include a separator char/string. Make sure that does not conflict with the options here.


def get_free_space_mb(dirname):
    """Return folder/drive free space (in megabytes)."""
    if platform.system() == 'Windows':
        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(  # type: ignore
            ctypes.c_wchar_p(dirname),
            None,
            None,
            ctypes.pointer(free_bytes))
        return free_bytes.value / 1024 / 1024
    else:
        st = os.statvfs(dirname)
        return st.f_bavail * st.f_frsize / 1024 / 1024


RunlogAccessor = Callable[[], RunLog]
TagsAccessor = Callable[[], TagCollection]


class ArchiverTag(Tag):
    def __init__(self,
                 runlog_accessor: RunlogAccessor,
                 tags_accessor: TagsAccessor,
                 data_log_interval_seconds: float) -> None:
        super().__init__("Archive filename")
        self.runlog_accessor = runlog_accessor
        self.tags_accessor = tags_accessor
        self.tags: TagCollection | None = None
        path = os.path.dirname(os.path.realpath(__file__))
        self.data_path = os.path.join(path, "data")
        self.last_save_tick: float = 0.0
        self.data_log_interval_seconds = data_log_interval_seconds
        self.file_path: str | None = None
        self.file_ready = False

        if not os.path.exists(self.data_path):
            os.makedirs(self.data_path)
            logger.info("Created archive directory: " + self.data_path)

        self.check_diskspace()

    def archive(self) -> str | None:
        return None

    def check_diskspace(self) -> bool:
        diskspace_free = get_free_space_mb(self.data_path)
        if diskspace_free < VERY_LOW_DISKSPACE_MB:
            logger.error("Drive is VERY low on diskspace.")
            return False
        elif diskspace_free < LOW_DISKSPACE_MB:
            logger.warning("Drive is low on diskspace. Archiver may soon fill up the drive.")
        return True

    def prepare_tags_file(self):
        """ Create file and write header row"""
        assert self.file_path is not None
        assert self.tags is not None
        with open(self.file_path, 'xt', newline='', encoding=encoding) as f:
            writer = csv.writer(f, delimiter=delimiter, quoting=quoting, escapechar=escapechar)
            tag_values = [tag.archive() for tag in self.tags]
            writer.writerow(
                ['Datetime (UTC)',] +
                [f'{tag.name} [{tag.unit}]' if tag.unit is not None else tag.name
                 for tag, val in zip(self.tags, tag_values)
                 if val is not None]
            )
        self.file_ready = True

    def write_tags_row(self):
        assert self.tags is not None
        if self.file_ready:
            assert self.file_path is not None
            with open(self.file_path, 'at', newline='', encoding=encoding) as f:
                writer = csv.writer(f, delimiter=delimiter, quoting=quoting, escapechar=escapechar)
                tag_values = [tag.archive() for tag in self.tags]
                utc_now = datetime.now(UTC)
                row = [str(utc_now)] + [val for val in tag_values if val is not None]
                try:
                    writer.writerow(row)
                except Exception:
                    logger.error(f"Error writing row: {row}", exc_info=True)

    def write_runlog(self):
        date_part = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        runlog_file_path = os.path.join(self.data_path, "archiver-runlog-" + date_part + ".txt")

        logger.info(f"Writing runlog to {runlog_file_path}")
        with open(runlog_file_path, 'xt', newline='', encoding=encoding) as f:
            writer = csv.writer(f, delimiter=delimiter, quoting=quoting)
            writer.writerow(['Command', 'Start time (UTC)', 'Completed time (UTC)'])
            try:
                runlog = self.runlog_accessor()
            except Exception:
                logger.error("Failed to access the runlog", exc_info=True)
                return

            for x in runlog.items:
                start = datetime.fromtimestamp(x.start, UTC)
                end = datetime.fromtimestamp(x.end, UTC) if x.end is not None else ""
                writer.writerow([x.name, start, end])

    def on_start(self, run_id: str):
        self.tags = self.tags_accessor()
        tick_time = time.time()
        date_part = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        filename = "archiver-" + date_part + ".txt"
        self.set_value(filename, tick_time)
        self.file_path = os.path.join(self.data_path, filename)
        if self.check_diskspace():
            logger.info("Archiver started using logfile " + self.file_path)
            self.prepare_tags_file()
        else:
            logger.error("Archiver will not run due to low diskspace")

    def on_tick(self, tick_time: float, increment_time: float):
        now = time.time()
        is_row_due = self.last_save_tick == 0 or self.last_save_tick + self.data_log_interval_seconds < now
        if is_row_due:
            self.write_tags_row()
            self.last_save_tick = now

    def on_stop(self):
        tick_time = time.time()
        logger.info("Stopped")
        self.set_value(None, tick_time)
        self.file_path = None
        self.file_ready = False
        self.write_runlog()




from time import time
from openpectus.engine.uod_builder_api import (
    UnitOperationDefinitionBase, UodCommand, UodBuilder,
    tags, RegisterDirection,
    PlotConfiguration, SubPlot, PlotAxis, PlotColorRegion
)


def create() -> UnitOperationDefinitionBase:
    builder = UodBuilder()

    def get_plot_configuration() -> PlotConfiguration:
        return PlotConfiguration(
            color_regions=[
                PlotColorRegion(
                    process_value_name='Category',
                    value_color_map={'Rising': '#66ff6633', 'Falling': '#6666ff33'})
            ],
            sub_plots=[SubPlot(
                axes=[PlotAxis(
                    label='FT01',
                    process_value_names=['FT01'],
                    y_max=20,
                    y_min=0,
                    color='#ff0000',
                ), PlotAxis(
                    label='FT02',
                    process_value_names=['FT02'],
                    y_max=20,
                    y_min=0,
                    color='#0000ff',
                )],
                ratio=1
            )],
            process_value_names_to_annotate=['Category'],
            x_axis_process_value_names=['Time']
        )

    def reset(cmd: UodCommand, value):
        cmd.context.tags["RESET"].set_value(1.0, time())

        # we don't support this. could be nice, though
        # for i in range(10):
        #     cmd.set_progress(.1 * i)
        #     yield

        cmd.set_complete()

        # TODO consider reset behavior:
        # Server reads the RESET register and executes its reset if non-zero and immideately sets the register value
        # bath to zero. This does not really match any of our register directions.

        # This does work but causes an extra reset from server
        # if cmd.get_iteration_count() > 0:
        #     cmd.set_complete()

    return (
        builder
        .with_instrument("Opcua-Test")
        .with_author("Demo Author", "demo@openpectus.org")
        .with_filename(__file__)
        .with_hardware_opcua(host="opc.tcp://127.0.0.1:48400")
        .with_location("Test location")
        .with_hardware_register("RESET", RegisterDirection.Both, path='Objects/2:System/2:RESET')
        .with_hardware_register("VA01", RegisterDirection.Read, path='Objects/2:System/2:VA01')
        .with_hardware_register("VA02", RegisterDirection.Read, path='Objects/2:System/2:VA02')
        .with_hardware_register("TT01", RegisterDirection.Read, path='Objects/2:System/2:TT01')
        .with_hardware_register("FT01", RegisterDirection.Read, path='Objects/2:System/2:FT01')
        .with_hardware_register("PU01", RegisterDirection.Read, path='Objects/2:System/2:PU01')

        .with_tag(tags.ReadingTag(name="RESET"))
        .with_tag(tags.ReadingTag(name="VA01"))
        .with_tag(tags.ReadingTag(name="VA02"))
        .with_tag(tags.ReadingTag(name="TT01"))  # unit="degC"
        .with_tag(tags.ReadingTag(name="FT01"))
        .with_tag(tags.ReadingTag(name="PU01"))

        .with_command("RESET", exec_fn=reset)

        .with_process_value("VA01")
        .with_process_value("VA02")
        .with_process_value("TT01")
        .with_process_value("FT01")
        .with_process_value("PU01")
        .with_process_value_choice(tag_name="RESET", command_options={"Reset": "RESET"})

        .with_plot_configuration(get_plot_configuration())
        .build()
    )



from types import NoneType
from typing import Any

import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
import openpectus.protocol.messages as M
from openpectus.protocol.exceptions import ProtocolDeserializationException

_message_namespaces = [AM, EM, M]
_message_namespace_names = [ns.__name__ for ns in _message_namespaces]


def serialize(msg: M.MessageBase) -> dict[str, Any]:
    """ Serialize a protocol message in a round-trippable fashion. """
    json_dict = msg.model_dump()
    json_dict["_type"] = type(msg).__qualname__
    json_dict["_ns"] = msg.__module__
    return json_dict


def deserialize(json_dict: dict[str, Any]) -> M.MessageBase:
    """ Deserializes a protocol message to the proper concrete MessageBase subtype.

    Supports messages serialized to dict by the serialize() function.
    """
    try:
        if "_type" not in json_dict.keys() or "_ns" not in json_dict.keys():
            raise ValueError("Deserialization error. Key '_type' or '_ns' missing.")
        message_type, module_name = json_dict["_type"], json_dict["_ns"]
        if module_name not in _message_namespace_names:
            raise ValueError(f"Message module name '{module_name}' is not a valid protocol message namespace.")
        ns = _message_namespaces[_message_namespace_names.index(module_name)]
        cls = getattr(ns, message_type, None)
        assert not isinstance(cls, NoneType), f"Class name {cls} is not defined in namespace '{module_name}'."
        msg = cls(**json_dict)
        assert isinstance(msg, M.MessageBase)
        return msg
    except Exception as ex:
        raise ProtocolDeserializationException(f"Deserialization protocol error: {ex}")



import openpectus.aggregator.models as Mdl
import openpectus.aggregator.routers.dto as Dto
import openpectus.protocol.aggregator_messages as AM


def parse_as_message(cmd: Dto.ExecutableCommand, readings: list[Mdl.ReadingInfo]) -> AM.AggregatorMessage:
    if cmd.command is None or cmd.command.strip() == "":
        raise ValueError("Command is empty")

    lines = cmd.command.splitlines(keepends=False)
    line_count = len(lines)
    if line_count < 1:
        raise ValueError("Command is empty")

    if cmd.source == Dto.CommandSource.UNIT_BUTTON:
        if line_count > 1:
            raise ValueError("Unit Button command must be a single line command")
        code = lines[0]
        if not code.istitle():
            raise ValueError("Unit Button commands must be title cased")
        return AM.InvokeCommandMsg(name=code)

    return _parse_advanced_command(cmd, readings)


def _parse_advanced_command(cmd: Dto.ExecutableCommand, readings: list[Mdl.ReadingInfo]) -> AM.AggregatorMessage:
    if cmd.value is None:
        return AM.InjectCodeMsg(pcode=cmd.command)

    if isinstance(cmd.value, Dto.ProcessValueCommandNumberValue):
        unit_str = ""
        if cmd.value.value_unit is not None and cmd.value.value_unit.strip() != "":
            unit_str = " " + cmd.value.value_unit.strip()
        return AM.InjectCodeMsg(pcode=f"{cmd.command}: {cmd.value.value}{unit_str}")

    elif isinstance(cmd.value, Dto.ProcessValueCommandFreeTextValue):
        return AM.InjectCodeMsg(pcode=f"{cmd.command}: {cmd.value.value}")

    elif isinstance(cmd.value, Dto.ProcessValueCommandChoiceValue):
        if cmd.command_id is None or cmd.command_id.strip() == "":
            raise ValueError("Invalid command. command_id empty")
        cmd_reading = next((r for r in readings if r.has_command_id(cmd.command_id)), None)
        if cmd_reading is None:
            raise ValueError(f"Invalid command. command_id '{cmd.command_id}' does not match a reading command")

        if cmd_reading.command_options is None:
            raise ValueError("Invalid command. Matching reading has no command_options")
        command = cmd_reading.command_options.get(cmd.value.value)
        if command is None:
            raise ValueError(f"Invalid command. Matching reading has no command_option with name {cmd.value.value}")

        return AM.InjectCodeMsg(pcode=command)

    raise ValueError(f"Command not supported: '{cmd}'")


def create_reading_commands(tag: Mdl.TagValue, reading: Mdl.ReadingInfo) -> list[Dto.ProcessValueCommand]:
    commands: list[Dto.ProcessValueCommand] = []
    for reading_cmd in reading.commands:
        if reading.discriminator == "reading_with_entry":
            command = Dto.ProcessValueCommand(
                command_id=reading_cmd.command_id,
                name=reading_cmd.name,
                command=reading_cmd.command,
                disabled=None,
                value=None
            )
            if reading.entry_data_type is None or reading.entry_data_type == "auto":
                raise ValueError(
                    f"Error in uod. Failed to build commands for ReadingWithEntry, tag: '{reading.tag_name}'." +
                    f"entry_data_type is {reading.entry_data_type}. Expected str/int/float")
            elif reading.entry_data_type == "str":
                command.value = Dto.ProcessValueCommandFreeTextValue(
                    value="",
                    value_type=Dto.ProcessValueType.STRING
                )
            elif reading.entry_data_type == "float":
                command.value = Dto.ProcessValueCommandNumberValue(
                    value=tag.value if isinstance(tag.value, (float, int)) else 0,
                    value_type=Dto.ProcessValueType.FLOAT,
                    value_unit=tag.value_unit,
                    valid_value_units=reading.valid_value_units
                )
            elif reading.entry_data_type == "int":
                command.value = Dto.ProcessValueCommandNumberValue(
                    value=int(tag.value) if isinstance(tag.value, (float, int)) else 0,
                    value_type=Dto.ProcessValueType.INT,
                    value_unit=tag.value_unit,
                    valid_value_units=reading.valid_value_units
                )
            commands.append(command)

        elif reading.discriminator == "reading_with_choice":
            if reading_cmd.choice_names is None or len(reading_cmd.choice_names) == 0:
                raise ValueError(
                    f"Error in uod. Failed to build commands for ReadingWithChoice, tag: '{reading.tag_name}'. " +
                    "Reading choice_names is empty")
            current_value = str(tag.value)
            command = Dto.ProcessValueCommand(
                command_id=reading_cmd.command_id,
                name=reading_cmd.name,
                command=reading_cmd.command,
                disabled=None,
                value=Dto.ProcessValueCommandChoiceValue(
                    value=current_value,
                    value_type=Dto.ProcessValueType.CHOICE,
                    options=[choice for choice in reading_cmd.choice_names]
                )
            )
            commands.append(command)

        else:
            raise ValueError(
                f"Error in uod. Failed to build commands for reading, tag: '{reading.tag_name}'. " +
                f"Unknown discriminator: {reading.discriminator}")

    return commands


def create_command_examples(commands: list[Mdl.CommandInfo]) -> list[Dto.CommandExample]:
    # TODO We probably need a headline and spacer elements
    examples: list[Dto.CommandExample] = []
    for cmd in commands:
        example = "No command information available"
        if cmd.docstring is not None and cmd.docstring != "":
            example = cmd.docstring
        examples.append(Dto.CommandExample(name=cmd.name, example=example))
    return examples



import unittest

import openpectus.aggregator.models as Mdl
import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
from openpectus.protocol.serialization import serialize, deserialize


class SerializationTest(unittest.TestCase):
    def test_serialization_RegisterEngineMsg(self):
        reg = EM.RegisterEngineMsg(
            computer_name="foo",
            uod_name="bar",
            uod_author_name="author_name",
            uod_author_email="",
            uod_filename="",
            location="baz",
            engine_version='0.0.1')
        reg_s = serialize(reg)
        self.assertEqual(
            {'_type': 'RegisterEngineMsg', '_ns': 'openpectus.protocol.engine_messages',
             'computer_name': 'foo', 'engine_version': '0.0.1', 'uod_name': 'bar',
             'uod_author_name': 'author_name', 'uod_author_email': '', 'uod_filename': '',
             'location': 'baz', 'secret': '',
             'sequence_number': -2, 'version': 0},
            reg_s)
        self.assertIsNotNone(reg_s)

    def test_round_trip_RegisterEngineMsg(self):
        reg = EM.RegisterEngineMsg(
            computer_name="foo",
            uod_name="bar",
            uod_author_name="author_name",
            uod_author_email="",
            uod_filename="",
            location="baz",
            engine_version='0.0.1')
        reg_s = serialize(reg)
        reg_d = deserialize(reg_s)
        self.assertIsNotNone(reg_d)
        assert isinstance(reg_d, EM.RegisterEngineMsg)
        self.assertEqual(reg.computer_name, reg_d.computer_name)
        self.assertEqual(reg.uod_name, reg_d.uod_name)

    def test_round_trip_AM_MethodMsg(self):
        am_method = AM.MethodMsg(method=Mdl.Method(lines=[Mdl.MethodLine(id='87', content='Foo')]))
        reg_s = serialize(am_method)
        reg_d = deserialize(reg_s)
        self.assertIsNotNone(reg_d)
        assert isinstance(reg_d, AM.MethodMsg)
        self.assertEqual(am_method.method, reg_d.method)

    def test_serialization_TagsUpdatedMsg(self):
        tu = EM.TagsUpdatedMsg(
            tags=[Mdl.TagValue(name="foo", tick_time=0, value="bar", value_formatted=None, value_unit="m")])
        tu_s = serialize(tu)
        self.assertIsNotNone(tu_s)

    def test_round_trip_TagsUpdatedMsg(self):
        tu = EM.TagsUpdatedMsg(
            tags=[Mdl.TagValue(name="foo", tick_time=0, value="bar", value_formatted=None, value_unit=None)]
        )
        tu_s = serialize(tu)
        self.assertIsNotNone(tu_s)

        tu_d = deserialize(tu_s)
        self.assertIsNotNone(tu_d)
        assert isinstance(tu_d, EM.TagsUpdatedMsg)
        self.assertEqual(tu_d.tags[0].name, tu.tags[0].name)


if __name__ == "__main__":
    unittest.main()



from datetime import datetime, timezone
import unittest
from unittest.mock import Mock, AsyncMock

from openpectus.aggregator.data import database
import openpectus.aggregator.models as Mdl
import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
from fastapi_websocket_rpc.schemas import RpcResponse
from openpectus.aggregator.aggregator import Aggregator
from openpectus.aggregator.aggregator_message_handlers import AggregatorMessageHandlers
from openpectus.protocol.aggregator_dispatcher import AggregatorDispatcher
import openpectus.aggregator.data.models as DMdl
from openpectus.protocol.models import SystemTagName


class AggregatorTest(unittest.IsolatedAsyncioTestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)

    async def create_channel_mock(self, engine_id: str | None):
        response = RpcResponse[str | None](result=engine_id, result_type=None)
        return Mock(close=AsyncMock(), other=Mock(get_engine_id_async=AsyncMock(return_value=response)))

    async def connectRpc(self, dispatcher: AggregatorDispatcher, engine_id: str | None):
        channel = await self.create_channel_mock(engine_id)
        await dispatcher._on_delayed_client_connect(channel)
        return channel

    async def disconnectRpc(self, dispatcher: AggregatorDispatcher, engine_id: str):
        channel = dispatcher._engine_id_channel_map[engine_id]
        await dispatcher.on_client_disconnect(channel)

    def createPublisherMock(self):
        return Mock(
            publish_process_units_changed=AsyncMock(),
            publish_control_state_changed=AsyncMock(),
        )

    async def test_register_engine(self):
        # setup in-memory database
        database.configure_db("sqlite:///:memory:")
        DMdl.DBModel.metadata.create_all(database._engine)  # type: ignore

        dispatcher = AggregatorDispatcher()
        aggregator = Aggregator(dispatcher, self.createPublisherMock())
        _ = AggregatorMessageHandlers(aggregator)
        register_engine_msg = EM.RegisterEngineMsg(
            computer_name='computer-name',
            uod_name='uod-name',
            uod_author_name="uod-author-name",
            uod_author_email="uod-author-email",
            uod_filename="uod-filename",
            location="test location",
            engine_version='0.0.1')
        engine_id = aggregator.create_engine_id(register_engine_msg)

        # connecting rpc with no response for engine id should close connection
        channel = await self.connectRpc(dispatcher, None)
        channel.close.assert_called()

        # registering while not registered before should succceed
        assert dispatcher._register_handler is not None
        resultMessage = await dispatcher._register_handler(register_engine_msg)
        assert isinstance(resultMessage, AM.RegisterEngineReplyMsg)
        self.assertEqual(resultMessage.success, True)

        # connecting rpc now should not close connection
        channel = await self.connectRpc(dispatcher, engine_id)
        channel.close.assert_not_called()

        # registering while already registered and still connected should fail
        resultMessage = await dispatcher._register_handler(register_engine_msg)
        assert isinstance(resultMessage, AM.RegisterEngineReplyMsg)
        self.assertEqual(resultMessage.success, False)

        # setup in-memory database - required for engine_disconnected
        database.configure_db("sqlite:///:memory:")
        DMdl.DBModel.metadata.create_all(database._engine)  # type: ignore
        with database.create_scope():
            # registering while already registered, but after disconnect should succeed
            await self.disconnectRpc(dispatcher, engine_id)
            resultMessage = await dispatcher._register_handler(register_engine_msg)
            assert isinstance(resultMessage, AM.RegisterEngineReplyMsg)
            self.assertEqual(resultMessage.success, True)

    async def test_register_engine_different_name(self):
        # setup in-memory database
        database.configure_db("sqlite:///:memory:")
        DMdl.DBModel.metadata.create_all(database._engine)  # type: ignore

        dispatcher = AggregatorDispatcher()
        aggregator = Aggregator(dispatcher, self.createPublisherMock())
        messageHandlers = AggregatorMessageHandlers(aggregator)
        register_engine_msg = EM.RegisterEngineMsg(
            computer_name='computer-name',
            uod_name='uod-name',
            uod_author_name="uod-author-name",
            uod_author_email="uod-author-email",
            uod_filename="uod-filename",
            location="test-loc",
            engine_version='0.0.1')
        register_engine_msg_different_computer = EM.RegisterEngineMsg(
            computer_name='computer-name2',
            uod_name='uod-name',
            uod_author_name="uod-author-name",
            uod_author_email="uod-author-email",
            uod_filename="uod-filename",
            location="test-loc",
            engine_version='0.0.1')
        engine_id1 = aggregator.create_engine_id(register_engine_msg)
        _ = aggregator.create_engine_id(register_engine_msg_different_computer)

        # registering engine 1 while not registered before should succceed
        resultMessage = await messageHandlers.handle_RegisterEngineMsg(register_engine_msg)
        assert isinstance(resultMessage, AM.RegisterEngineReplyMsg)
        self.assertEqual(resultMessage.success, True)

        # connecting rpc now for registered engine should not close connection
        channel = await self.connectRpc(dispatcher, engine_id1)
        channel.close.assert_not_called()

        # registering while already registered with same name should fail
        resultMessage = await messageHandlers.handle_RegisterEngineMsg(register_engine_msg)
        assert isinstance(resultMessage, AM.RegisterEngineReplyMsg)
        self.assertEqual(resultMessage.success, False)

        # registering while already registered but with different name should succeed
        resultMessage = await messageHandlers.handle_RegisterEngineMsg(register_engine_msg_different_computer)
        assert isinstance(resultMessage, AM.RegisterEngineReplyMsg)
        self.assertEqual(resultMessage.success, True)

class AggregatorEventsTest(unittest.IsolatedAsyncioTestCase):

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.stored_tags = []
        self.plot_log_repo = Mock(store_tag_values=self.store_tag_values)
        self.aggregator = Aggregator(Mock(), Mock())
        self.engine_data = Mdl.EngineData(
            engine_id="test_engine", computer_name="", engine_version="", hardware_str="",
            uod_name="", uod_author_name="", uod_author_email="", uod_filename="", location="",
            data_log_interval_seconds=5
        )

    def createTag(self, name: str, tick: float, value: str):
        return Mdl.TagValue(name=name, tick_time=tick, value=value, value_formatted=None, value_unit=None)

    def store_tag_values(self, engine_id: str, run_id: str, tags: list[Mdl.TagValue]):
        self.stored_tags.extend(tags)

    def process_tags(self, tags: list[Mdl.TagValue]):
        for tag in tags:
            self.engine_data.tags_info.upsert(tag)


    async def test_persist_tag_values_can_save_new_tags(self):
        run_id = self.createTag(SystemTagName.RUN_ID.value, 1, "run1")
        tags = [
            run_id,
            self.createTag("a", 1, "v1"),
            self.createTag("b", 1, "v1"),
        ]
        self.process_tags(tags)

        # create run_data to simulate a started run
        self.engine_data.run_data = Mdl.RunData.empty(run_id="run1", run_started=datetime.now(timezone.utc))

        self.aggregator.from_engine._persist_tag_values(self.engine_data, self.plot_log_repo)

        self.assertEqual(tags, self.stored_tags)

    async def test_persist_tag_values_can_update_tag_with_newer_after_threshold(self):
        run_id = self.createTag(SystemTagName.RUN_ID.value, 1, "run1")

        tags = [
            run_id,
            self.createTag("a", 1, "v1"),
            self.createTag("b", 1, "v1"),
        ]
        self.process_tags(tags)

        # create run_data to simulate a started run
        self.engine_data.run_data = Mdl.RunData.empty(run_id="run1", run_started=datetime.now(timezone.utc))

        self.aggregator.from_engine._persist_tag_values(self.engine_data, self.plot_log_repo)
        self.stored_tags.clear()

        # data newer but not beyond threshold is not saved
        tmp_tags = [
            self.createTag("a", 1.1 + 3, "v2"),  # threshold is 5 seconds
        ]
        self.process_tags(tmp_tags)
        self.aggregator.from_engine._persist_tag_values(self.engine_data, self.plot_log_repo)

        self.assertEqual([], self.stored_tags)

        # data newer and beyond threshold is saved
        new_tags = [
            self.createTag("a", 1.1 + 5, "v2"),  # threshold is 5 seconds
        ]
        self.process_tags(new_tags)
        self.aggregator.from_engine._persist_tag_values(self.engine_data, self.plot_log_repo)

        self.assertEqual(new_tags, self.stored_tags)

    async def test_persist_tag_values_cannot_update_tag_with_older(self):
        run_id = self.createTag(SystemTagName.RUN_ID.value, 1, "run1")
        tags = [
            run_id,
            self.createTag("a", 2, "v1"),
            self.createTag("b", 2, "v1"),
        ]
        self.process_tags(tags)

        # create run_data to simulate a started run
        self.engine_data.run_data = Mdl.RunData.empty(run_id="run1", run_started=datetime.now(timezone.utc))

        self.aggregator.from_engine._persist_tag_values(self.engine_data, self.plot_log_repo)
        self.stored_tags.clear()
        self.assertEqual(2, self.engine_data.run_data.latest_persisted_tick_time)

        new_tags = [
            self.createTag("a", 1, "v2"),
        ]

        # So we allow this to update the tag value with an older time and give a warning
        # - but we disallow persisting this older value??
        self.process_tags(new_tags)

        self.aggregator.from_engine._persist_tag_values(self.engine_data, self.plot_log_repo)

        self.assertEqual([], self.stored_tags)


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
from datetime import datetime, timezone

import openpectus.aggregator.models as Mdl
import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
import openpectus.protocol.messages as M
from openpectus.aggregator.data import database
from openpectus.aggregator.data.repository import RecentRunRepository, PlotLogRepository, RecentEngineRepository
from openpectus.aggregator.frontend_publisher import FrontendPublisher
from openpectus.aggregator.models import EngineData
from openpectus.protocol.aggregator_dispatcher import AggregatorDispatcher
from openpectus.protocol.models import SystemTagName, MethodStatusEnum

logger = logging.getLogger(__name__)

EngineDataMap = dict[str, EngineData]


class FromEngine:
    def __init__(self, engine_data_map: EngineDataMap, publisher: FrontendPublisher):
        self._engine_data_map = engine_data_map
        self.publisher = publisher

    def register_engine_data(self, engine_data: EngineData):
        engine_id = engine_data.engine_id
        logger.debug(f"Data for engine {engine_id} registered")
        self._engine_data_map[engine_id] = engine_data

        self._try_restore_reconnected_engine_data(engine_data)

        asyncio.create_task(self.publisher.publish_process_units_changed())
        asyncio.create_task(self.publisher.publish_control_state_changed(engine_data.engine_id))

    def _try_restore_reconnected_engine_data(self, engine_data: EngineData):
        engine_id = engine_data.engine_id
        logger.debug(f"Trying to restore engine data for recent engine: {engine_id}")
        with database.create_scope():
            repo = RecentEngineRepository(database.scoped_session())
            recent_engine = repo.get_recent_engine_by_engine_id(engine_id)
        if recent_engine is not None:  # reconnecting an existing engine
            if recent_engine.run_id is not None:  # that was in an active run when disconnected
                run_id = recent_engine.run_id
                logger.debug(f"Applying run_data {run_id=} from recent_engine")
                if recent_engine.run_started is None:
                    logger.warning("Recent engine had a run without run_started value. Using now as run_started")
                    run_started = datetime.now(timezone.utc)
                else:
                    run_started = recent_engine.run_started
                engine_data.run_data = Mdl.RunData.empty(run_id=run_id, run_started=run_started)
                engine_data.contributors = set(recent_engine.contributors)
            else:
                logger.info("Recent engine data has no active run")
        else:
            logger.info("No recent engine data found")

    def engine_connected(self, engine_id: str):
        logger.debug("engine_connected")

    def engine_disconnected(self, engine_id: str):
        logger.debug("engine_disconnected")
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is not None:
            with database.create_scope():
                repo = RecentEngineRepository(database.scoped_session())
                repo.store_recent_engine(engine_data)
            logger.info("Recent engine saved")
            del self._engine_data_map[engine_id]
            asyncio.create_task(self.publisher.publish_process_units_changed())
        else:
            logger.warning("No data to save for engine " + engine_id)

    def run_started(self, msg: EM.RunStartedMsg):
        engine_id = msg.engine_id
        engine_data = self._engine_data_map.get(engine_id)
        with database.create_scope():
            if engine_data is None:
                logger.error("No engine data available on run_started for engine " + engine_id)
                return
            if not engine_data.has_run():
                engine_data.reset_run()
                engine_data.run_data = Mdl.RunData.empty(
                    run_id=msg.run_id,
                    run_started=datetime.fromtimestamp(msg.started_tick, timezone.utc)
                )
            elif engine_data.run_data.run_id == msg.run_id:
                # be idempotent and just accept this duplicate message
                logger.warning("Event run_started occurred with same id as the current run. Ignoring")
            else:
                _run_id = engine_data.run_data.run_id
                logger.error(f"Current run_id {_run_id} does not match RunStartedMsg run_id {msg.run_id}")
                recent_run_repo = RecentRunRepository(database.scoped_session())
                try:
                    recent_run_repo.store_recent_run(engine_data)
                    logger.info(f"Stopping existing run and store it as recent run {_run_id=}")
                    engine_data.reset_run()
                except Exception:
                    logger.error(f"Failed to persist recent run {_run_id=}", exc_info=True)

            plot_log_repo = PlotLogRepository(database.scoped_session())
            plot_log_repo.create_plot_log(engine_data, msg.run_id)

        asyncio.create_task(self.publisher.publish_control_state_changed(engine_id))
        logger.info(f"Run {msg.run_id} started")

    def run_stopped(self, msg: EM.RunStoppedMsg):
        engine_id = msg.engine_id
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is None:
            logger.error("No engine data available on run_stopped for engine " + engine_id)
            return
        if not engine_data.has_run():
            logger.warning("No engine run_data available on run_stopped for engine " + engine_id)
            return

        with database.create_scope():
            recent_run_repo = RecentRunRepository(database.scoped_session())
            _run_id = engine_data.run_data.run_id
            if _run_id != msg.run_id:
                logger.error(
                    "Run_id mismatch in run_stopped. " +
                    f"engine {engine_id}, run_data run_id: {engine_data.run_data.run_id}, message run_id: {_run_id}")
                logger.warning(f"Saving existing run {_run_id}. No data is available for the other run")
                try:
                    recent_run_repo.store_recent_run(engine_data)
                    logger.info(f"Stored recent run {_run_id=}")
                except Exception:
                    logger.error(f"Failed to persist recent run {_run_id=}")
                engine_data.reset_run()
            else:
                engine_data.run_data.runlog = msg.runlog
                engine_data.method = msg.method
                try:
                    recent_run_repo.store_recent_run(engine_data)
                    logger.info(f"Stored recent run {_run_id=}")
                except Exception:
                    logger.error(f"Failed to persist recent run {_run_id=}")

        # clear current run_data
        engine_data.reset_run()
        asyncio.create_task(self.publisher.publish_control_state_changed(engine_id))
        asyncio.create_task(self.publisher.publish_method_state_changed(engine_id))
        asyncio.create_task(self.publisher.publish_run_log_changed(engine_id))
        logger.info(f"Run {msg.run_id} stopped")


    def uod_info_changed(
            self,
            engine_id: str,
            readings: list[Mdl.ReadingInfo],
            commands: list[Mdl.CommandInfo],
            plot_configuration: Mdl.PlotConfiguration,
            hardware_str: str,
            required_roles: set[str],
            data_log_interval_seconds: float):
        try:
            self._engine_data_map[engine_id].readings = readings
            self._engine_data_map[engine_id].commands = commands
            self._engine_data_map[engine_id].plot_configuration = plot_configuration
            self._engine_data_map[engine_id].hardware_str = hardware_str
            self._engine_data_map[engine_id].required_roles = required_roles
            self._engine_data_map[engine_id].data_log_interval_seconds = data_log_interval_seconds
        except KeyError:
            logger.error(f'No engine registered under id {engine_id} when trying to set uod info.')

    def tag_values_changed(self, engine_id: str, changed_tag_values: list[Mdl.TagValue]):
        with database.create_scope():
            plot_log_repo = PlotLogRepository(database.scoped_session())

            engine_data = self._engine_data_map.get(engine_id)
            if engine_data is None:
                logger.error(f'No engine registered under id {engine_id} when trying to upsert tag values.')
                return

            for changed_tag_value in changed_tag_values:
                if changed_tag_value.name == SystemTagName.METHOD_STATUS.value and engine_data.has_run():
                    if changed_tag_value.value == MethodStatusEnum.ERROR:
                        engine_data.run_data.interrupted_by_error = True
                    else:
                        engine_data.run_data.interrupted_by_error = False

                was_inserted = engine_data.tags_info.upsert(changed_tag_value)

                if changed_tag_value.name in [
                    SystemTagName.METHOD_STATUS.value,
                    SystemTagName.SYSTEM_STATE.value,
                    SystemTagName.RUN_ID.value,
                ]:
                    asyncio.create_task(self.publisher.publish_process_units_changed())

                # if a tag doesn't appear with value until after start and run_id, we need to store the info here
                if was_inserted and engine_data.has_run():
                    plot_log_repo.store_new_tag_info(engine_data.engine_id, engine_data.run_data.run_id, changed_tag_value)

            self._persist_tag_values(engine_data, plot_log_repo)

    def __run_id_changed(
            self,
            plot_log_repo: PlotLogRepository,
            recent_run_repo: RecentRunRepository,
            engine_data: EngineData,
            run_id_tag: Mdl.TagValue):
        """ Handles persistance related to start and end of a run """
        raise NotImplementedError("not needed anymore")

        logger.info(f"RunId changed from {engine_data.run_id} to {run_id_tag.value}, Engine: {engine_data.engine_id}")
        if run_id_tag.value is None and engine_data.run_id is None:
            logger.info("Engine has no active run")
        elif run_id_tag.value is None and engine_data.run_id is not None:
            # Run stopped
            logger.info(f"Run was stopped. Saving Recent Run. Engine: {engine_data.engine_id}")
            recent_run_repo.store_recent_run(engine_data)
            engine_data.run_data.error_log = Mdl.AggregatedErrorLog.empty()
            asyncio.create_task(self.publisher.publish_error_log_changed(engine_data.engine_id))
        elif run_id_tag.value is not None and engine_data.run_id is not None:
            # Ongoing run. run_id was only detected as changed because of complete tags set,
            # caused by e.g. an aggregator restart
            if run_id_tag.value != engine_data.run_id:
                logger.error(f"Run_Id inconsistent, run_id tag {run_id_tag.value}, engine_data {engine_data.run_id}")
            else:
                logger.info(f"Run {run_id_tag.value} resumed")
        else:
            # Run started
            logger.info(f"A new Run was started, Engine: {engine_data.engine_id}, Run_Id: {str(run_id_tag.value)}")
            engine_data.run_data.run_started = datetime.fromtimestamp(run_id_tag.tick_time, timezone.utc)
            plot_log_repo.create_plot_log(engine_data, str(run_id_tag.value))

    def _persist_tag_values(self, engine_data: EngineData, plot_log_repo: PlotLogRepository):
        if not engine_data.has_run():
            return
        assert engine_data.run_data is not None  # would be nice to have this given by the has_run() result

        latest_persisted_tick_time = engine_data.run_data.latest_persisted_tick_time
        tag_values = engine_data.tags_info.map.values()
        latest_tag_tick_time = max([tag.tick_time for tag in tag_values]) if len(tag_values) > 0 else 0
        time_threshold_exceeded = latest_persisted_tick_time is None\
            or latest_tag_tick_time - latest_persisted_tick_time > engine_data.data_log_interval_seconds

        if engine_data.run_data.run_id is not None and time_threshold_exceeded:
            tag_values_to_persist = [tag_value.model_copy() for tag_value in engine_data.tags_info.map.values()
                                     if latest_persisted_tick_time is None
                                     or tag_value.tick_time > latest_persisted_tick_time]
            """
            We manipulate the tick_time of the tagValues we persist. But it's not changing it to something that didn't
            exist in the engine, because we change the tick_time to a tick_time that comes from an actual later reading
            where those tagValues we manipulate was read and simply had not changed value since the value we have been
            reported.
            It's difficult to explain, but after this manipulation, the tagValues will still match an actual tagValue
            read in the engine, just not one reported.
            We do this because it solves a lot of issues when we later try to match values based on the tick_time.
            """
            highest_tick_time_to_persist = max([tag_Value.tick_time for tag_Value in tag_values_to_persist])
            for tag_value_to_persist in tag_values_to_persist:
                tag_value_to_persist.tick_time = highest_tick_time_to_persist

            # Note: to store tag values, the run_id is needed
            plot_log_repo.store_tag_values(engine_data.engine_id, engine_data.run_data.run_id, tag_values_to_persist)
            engine_data.run_data.latest_persisted_tick_time = highest_tick_time_to_persist
            engine_data.run_data.latest_tag_time = highest_tick_time_to_persist

    def runlog_changed(self, engine_id: str, run_id: str, runlog: Mdl.RunLog):
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is None:
            logger.error(f'No engine registered under id {engine_id} when trying to set run log.')
            return
        if not engine_data.has_run():
            logger.error(f"No run_data available for engine {engine_id}, can't set runlog")
            return
        if engine_data.run_data.run_id is None or engine_data.run_data.run_id != run_id:
            logger.error(
                "Run_id mismatch in runlog_changed. " +
                f"engine {engine_id}, run_data run_id: {engine_data.run_data.run_id}, message run_id: {run_id}")
            return

        if engine_data.run_data.runlog != runlog:
            engine_data.run_data.runlog = runlog
            asyncio.create_task(self.publisher.publish_run_log_changed(engine_id))

    def control_state_changed(self, engine_id: str, control_state: Mdl.ControlState):
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is None:
            logger.error(f'No engine registered under id {engine_id} when trying to set control state.')
            return
        if engine_data.control_state != control_state:
            engine_data.control_state = control_state
            asyncio.create_task(self.publisher.publish_control_state_changed(engine_id))

    def method_state_changed(self, engine_id: str, method_state: Mdl.MethodState):
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is None:
            logger.error(f'No engine registered under id {engine_id} when trying to set method state.')
            return
        if engine_data.method_state != method_state:
            engine_data.method_state = method_state
            asyncio.create_task(self.publisher.publish_method_state_changed(engine_id))

    def error_log_changed(self, engine_id: str, error_log: Mdl.ErrorLog):
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is None:
            logger.error(f'No engine registered under id {engine_id} when trying to set error log.')
            return
        engine_data.error_log.aggregate_with(error_log)
        asyncio.create_task(self.publisher.publish_error_log_changed(engine_id))


class FromFrontend:
    def __init__(self, engine_data_map: EngineDataMap, dispatcher: AggregatorDispatcher):
        self._engine_data_map = engine_data_map
        self.dispatcher = dispatcher

    async def method_saved(self, engine_id: str, method: Mdl.Method, user_name: str) -> bool:
        try:
            response = await self.dispatcher.rpc_call(engine_id, message=AM.MethodMsg(method=method))
            if isinstance(response, M.ErrorMessage):
                logger.error(f"Failed to set method. Engine response: {response.message}")
                return False
        except Exception:
            logger.error("Failed to set method", exc_info=True)
            return False

        # update local method state
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is not None:
            engine_data.method = method
            engine_data.contributors.add(user_name)
        return True

    async def request_cancel(self, engine_id, line_id: str, user_name: str) -> bool:
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is None:
            logger.warning(f"Cannot request cancel, engine {engine_id} not found")
            return False
        try:
            response = await self.dispatcher.rpc_call(engine_id, message=AM.CancelMsg(exec_id=line_id))
            if isinstance(response, M.ErrorMessage):
                logger.error(f"Cancel request failed. Engine response: {response.message}")
                return False
        except Exception:
            logger.error("Cancel request failed with exception", exc_info=True)
            return False
        engine_data.contributors.add(user_name)
        return True

    async def request_force(self, engine_id, line_id: str, user_name: str) -> bool:
        engine_data = self._engine_data_map.get(engine_id)
        if engine_data is None:
            logger.warning(f"Cannot request force, engine {engine_id} not found")
            return False
        try:
            response = await self.dispatcher.rpc_call(engine_id, message=AM.ForceMsg(exec_id=line_id))
            if isinstance(response, M.ErrorMessage):
                logger.error(f"Force request failed. Engine response: {response.message}")
                return False
        except Exception:
            logger.error("Force request failed with exception", exc_info=True)
            return False
        engine_data.contributors.add(user_name)
        return True


class Aggregator:
    def __init__(self, dispatcher: AggregatorDispatcher, publisher: FrontendPublisher, secret: str = "") -> None:
        self._engine_data_map: EngineDataMap = {}
        """ all client data except channels, indexed by engine_id """
        self.dispatcher = dispatcher
        self.from_frontend = FromFrontend(self._engine_data_map, dispatcher)
        self.from_engine = FromEngine(self._engine_data_map, publisher)
        self.secret = secret

    def create_engine_id(self, register_engine_msg: EM.RegisterEngineMsg):
        """ Defines the generation of the engine_id that is uniquely assigned to each engine.

        TODO: Considerations:
            - engine name should be machine name
            - uod hash should probably be included

        Implications of the registration process
        - historical data; we should not corrupt historical data by accidentially reusing engine_id
        - number of cards shown; we should not show many irrelevant cards in frontend of superseeded engine_ids
        """
        return register_engine_msg.computer_name + "_" + register_engine_msg.uod_name

    def get_registered_engine_data(self, engine_id: str):
        return self._engine_data_map.get(engine_id)

    def get_all_registered_engine_data(self) -> list[EngineData]:
        return list(self._engine_data_map.values())

    def has_registered_engine_id(self, engine_id: str) -> bool:
        return engine_id in self._engine_data_map.keys()



"""introduce required_roles

Revision ID: d53dfbe96f00
Revises: b792689d4cd4
Create Date: 2024-11-27 11:15:05.624346

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'd53dfbe96f00'
down_revision: Union[str, None] = 'b792689d4cd4'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentEngines', schema=None) as batch_op:
        batch_op.add_column(sa.Column('required_roles', sa.JSON(), nullable=False, server_default='[]'))

    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.add_column(sa.Column('required_roles', sa.JSON(), nullable=False, server_default='[]'))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.drop_column('required_roles')

    with op.batch_alter_table('RecentEngines', schema=None) as batch_op:
        batch_op.drop_column('required_roles')
    # ### end Alembic commands ###



import unittest

import openpectus.aggregator.models as Mdl
import openpectus.aggregator.routers.dto as Dto
from openpectus.lang.exec.readings import Reading, ReadingWithEntry, ReadingWithChoice
import openpectus.protocol.aggregator_messages as AM

from openpectus.aggregator import command_util as cu


def create_ReadingInfo(reading: Reading) -> Mdl.ReadingInfo:
    reading.build_commands_list()
    return reading.as_reading_info()


class CommantUtilTest(unittest.TestCase):

    def test_unit_button_parse(self):

        def test(command, name, expectSuccess=True, expectedResultType: type | None = None):
            with self.subTest(command + "_" + name):
                cmd = Dto.ExecutableCommand(
                    command_id=None,
                    command=command,
                    source=Dto.CommandSource.UNIT_BUTTON,
                    name=name,
                    value=None)

                if expectSuccess:
                    msg = cu.parse_as_message(cmd, [])
                    if expectedResultType is not None:
                        self.assertIsInstance(msg, expectedResultType)
                else:
                    with self.assertRaises(ValueError, msg="Expected parse error"):
                        cu.parse_as_message(cmd, [])

        test("Start", "", True, AM.InvokeCommandMsg)
        test("Stop", "", True, AM.InvokeCommandMsg)

        test("", "", False)
        test("stop", "", False)
        test("Stop\nStart", "", False)


    def test_cmd_without_value_parse_as_injected_with_command_as_pcode(self):
        cmd = Dto.ExecutableCommand(
            command_id=None,
            command="foo",
            source=Dto.CommandSource.PROCESS_VALUE,
            name=None,
            value=None)

        msg = cu.parse_as_message(cmd, [])
        assert isinstance(msg, AM.InjectCodeMsg)
        self.assertEqual("foo", msg.pcode)


    def test_cmd_with_value_Number_float_no_unit(self):
        reading = create_ReadingInfo(ReadingWithEntry(tag_name="foo", entry_data_type="float"))

        cmd = Dto.ExecutableCommand(
            command_id=reading.commands[0].command_id,
            command="foo",
            source=Dto.CommandSource.PROCESS_VALUE,
            name=None,
            value=Dto.ProcessValueCommandNumberValue(
                value=13.5,
                value_unit="",
                valid_value_units=[],
                value_type=Dto.ProcessValueType.FLOAT
            )
        )

        msg = cu.parse_as_message(cmd, [reading])

        assert isinstance(msg, AM.InjectCodeMsg)
        self.assertEqual("foo: 13.5", msg.pcode)

    def test_cmd_with_value_Number_int_unit(self):
        reading = create_ReadingInfo(ReadingWithEntry(tag_name="bar", entry_data_type="int"))

        cmd = Dto.ExecutableCommand(
            command_id=reading.commands[0].command_id,
            command="bar",
            source=Dto.CommandSource.PROCESS_VALUE,
            name=None,
            value=Dto.ProcessValueCommandNumberValue(
                value=87,
                value_unit="kg",
                valid_value_units=["g", "kg"],
                value_type=Dto.ProcessValueType.INT
            )
        )

        msg = cu.parse_as_message(cmd, [reading])

        assert isinstance(msg, AM.InjectCodeMsg)
        self.assertEqual("bar: 87 kg", msg.pcode)

    def test_cmd_with_value_FreeText_(self):
        reading = create_ReadingInfo(ReadingWithEntry(tag_name="bar", entry_data_type="str"))

        cmd = Dto.ExecutableCommand(
            command_id=reading.commands[0].command_id,
            command="bar",
            source=Dto.CommandSource.PROCESS_VALUE,
            name=None,
            value=Dto.ProcessValueCommandFreeTextValue(
                value="freetext VaLuE",
                value_type=Dto.ProcessValueType.STRING
            )
        )

        msg = cu.parse_as_message(cmd, [reading])

        assert isinstance(msg, AM.InjectCodeMsg)
        self.assertEqual("bar: freetext VaLuE", msg.pcode)

    def test_cmd_with_value_Choice_(self):
        reading = create_ReadingInfo(ReadingWithChoice(tag_name="foo", command_options={'A': 'Foo: A', 'B': 'Foo: B'}))

        cmd = Dto.ExecutableCommand(
            command_id=reading.commands[0].command_id,
            command="foo",
            source=Dto.CommandSource.PROCESS_VALUE,
            name=None,
            value=Dto.ProcessValueCommandChoiceValue(
                value="B",  # <- name of selected choice is provided here
                value_type=Dto.ProcessValueType.CHOICE,
                options=['A', 'B']
            ))

        msg = cu.parse_as_message(cmd, [reading])

        assert isinstance(msg, AM.InjectCodeMsg)
        self.assertEqual("Foo: B", msg.pcode)



from __future__ import annotations

import inspect
import logging
from enum import Enum
from typing import Generator, Iterable
from uuid import UUID

from openpectus.lang.exec.events import EventEmitter
import openpectus.lang.exec.units as units
from openpectus.lang.exec.base_unit import BaseUnitProvider
from openpectus.lang.exec.commands import InterpreterCommandEnum
from openpectus.lang.exec.errors import (
    EngineError, InterpretationError, InterpretationInternalError, NodeInterpretationError
)
from openpectus.lang.exec.runlog import RuntimeInfo, RuntimeRecordStateEnum
from openpectus.lang.exec.tags import (
    TagCollection, SystemTagName,
)
from openpectus.lang.model.pprogram import (
    PCommandWithDuration,
    PComment,
    PErrorInstruction,
    PInjectedNode,
    PNode,
    PProgram,
    PInstruction,
    PBlank,
    PBlock,
    PEndBlock,
    PEndBlocks,
    PWatch,
    PAlarm,
    PCommand,
    PMark,
)
from typing_extensions import override

logger = logging.getLogger(__name__)

term_uod = "Unit Operation Definition file."

class PNodeVisitor:
    def visit(self, node):
        method_name = 'visit_' + type(node).__name__
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        raise Exception('No visit_{} method'.format(type(node).__name__))


class ARType(Enum):
    PROGRAM = "PROGRAM",
    BLOCK = "BLOCK",
    WATCH = "WATCH",
    ALARM = "ALARM",
    INJECTED = "INJECTED",


class ActivationRecord:
    def __init__(
            self,
            owner: PNode,
            start_time: float = -1.0,
    ) -> None:
        self.owner: PNode = owner
        self.start_time: float = start_time
        self.complete: bool = False
        self.artype: ARType = self._get_artype(owner)


    def fill_start(self, start_time: float):
        self.start_time: float = start_time

    def _get_artype(self, node: PNode) -> ARType:
        if isinstance(node, PProgram):
            return ARType.PROGRAM
        elif isinstance(node, PBlock):
            return ARType.BLOCK
        elif isinstance(node, PWatch):
            return ARType.WATCH
        elif isinstance(node, PAlarm):
            return ARType.ALARM
        elif isinstance(node, PInjectedNode):
            return ARType.INJECTED
        else:
            raise NotImplementedError(f"ARType of node type '{type(node).__name__}' unknown")

    def __str__(self) -> str:
        return f"AR {self.owner} | {self.artype} | complete: {self.complete}"


class CallStack:
    def __init__(self):
        self._records = []

    @property
    def records(self) -> list[ActivationRecord]:
        return list(self._records)

    def iterate_from_top(self) -> Iterable[ActivationRecord]:
        for ar in reversed(self._records):
            yield ar

    def push(self, ar: ActivationRecord):
        self._records.append(ar)

    def pop(self) -> ActivationRecord:
        return self._records.pop()

    def peek(self) -> ActivationRecord:
        return self._records[-1]

    def __str__(self):
        s = '\n'.join(repr(ar) for ar in reversed(self._records))
        s = f'CALL STACK\n{s}\n\n'
        return s

    def __repr__(self):
        return self.__str__()


class LogEntry():
    def __init__(self, time: float, unit_time: float | None = None, message: str = '') -> None:
        self.time: float = time
        self.unit_time: float | None = unit_time
        self.message: str = message


class InterpreterContext():
    """ Defines the context of program interpretation"""

    @property
    def tags(self) -> TagCollection:
        raise NotImplementedError()

    def schedule_execution(self, name: str, exec_id: UUID | None = None, **kvargs):
        raise NotImplementedError()

    @property
    def emitter(self) -> EventEmitter:
        raise NotImplementedError()

    @property
    def base_unit_provider(self) -> BaseUnitProvider:
        raise NotImplementedError()


GenerationType = Generator[None, None, None] | None


class PInterpreter(PNodeVisitor):
    def __init__(self, program: PProgram, context: InterpreterContext) -> None:
        self._program = program
        self.context = context
        self.stack: CallStack = CallStack()
        self.interrupts: list[tuple[ActivationRecord, GenerationType]] = []
        self.running: bool = False

        self.start_time: float = 0
        self._tick_time: float = 0
        self._tick_number: int = -1

        self.process_instr: GenerationType = None

        self.runtimeinfo: RuntimeInfo = RuntimeInfo()
        logger.debug("Interpreter initialized")

    def get_marks(self) -> list[str]:
        records: list[tuple[str, int]] = []
        for r in self.runtimeinfo.records:
            if isinstance(r.node, PMark):
                completed_states = [st for st in r.states if st.state_name == RuntimeRecordStateEnum.Completed]
                for completed_state in completed_states:
                    end_tick = completed_state.state_tick
                    records.append((r.node.name, end_tick))

        def sort_fn(t: tuple[str, int]) -> int:
            return t[1]

        records.sort(key=sort_fn)
        return [t[0] for t in records]

    def inject_node(self, program: PProgram):
        """ Inject the child nodes of program into the running program in the current scope
        to be executed as the next instruction. """
        node = PInjectedNode(None)
        node.children = program.children
        ar = ActivationRecord(node, self._tick_time)
        self._register_interrupt(ar, self._create_interrupt_handler(node, ar))

    def _interpret(self) -> GenerationType:
        """ Create generator for interpreting the main program. """
        self.running = True
        tree = self._program
        if tree is None:
            return None
        tree.reset_runtime_state(recursive=True)
        logger.info("Reset runtime state")
        yield from self.visit(tree)


    def _register_interrupt(self, ar: ActivationRecord, handler: GenerationType):
        logger.debug(f"Interrupt handler registered for {ar}")
        self.interrupts.append((ar, handler))

    def _unregister_interrupt(self, ar: ActivationRecord):
        pairs = list([(x, y) for (x, y) in self.interrupts if x is ar])
        for pair in pairs:
            self.interrupts.remove(pair)
            logger.debug(f"Interrupt handler unregistered for {ar}")

    def _run_interrupt_handlers(self):
        instr_count = 0
        for ar, handler in list(self.interrupts):
            logger.debug(f"Handler count: {len(self.interrupts)}")
            node = ar.owner
            if isinstance(node, (PWatch, PInjectedNode, PAlarm)):
                logger.debug(f"Interrupt {str(node)}")
                self.stack.push(ar)
                try:
                    assert handler is not None, "Handler is None"
                    next(handler)
                    instr_count += 1
                except StopIteration:
                    pass
                finally:
                    self.stack.pop()

                if ar.complete:
                    self._unregister_interrupt(ar)
                    if isinstance(node, PAlarm):
                        node.reset_runtime_state(recursive=True)
                        ar = ActivationRecord(node)
                        self._register_interrupt(ar, self._create_interrupt_handler(node, ar))
            else:
                raise NotImplementedError(f"Interrupt for node type {str(node)} not implemented")
        return instr_count > 0

    def _create_interrupt_handler(self, node: PNode, ar: ActivationRecord) -> GenerationType:
        ar.fill_start(self._tick_time)
        yield from self.visit(node)

    def tick(self, tick_time: float, tick_number: int):
        self._tick_time = tick_time
        self._tick_number = tick_number

        logger.debug(f"Tick {self._tick_number}")

        if self.process_instr is None:
            self.process_instr = self._interpret()

        # execute one iteration of program
        assert self.process_instr is not None, "self.process_instr is None"
        try:
            next(self.process_instr)
        except StopIteration:
            pass
        except AssertionError as ae:
            raise InterpretationError(message=str(ae), exception=ae)
        except (InterpretationInternalError, InterpretationError):
            raise
        except EngineError:  # a method call on context failed - engine will know what to do
            raise
        except Exception as ex:
            logger.error("Unhandled interpretation error", exc_info=True)
            raise InterpretationError("Interpreter error") from ex

        # execute one iteration of each interrupt
        try:
            work_done = self._run_interrupt_handlers()
            if not work_done:
                pass
        except AssertionError as ae:
            raise InterpretationError(message=str(ae), exception=ae)
        except (InterpretationInternalError, InterpretationError):
            logger.error("Interpreter error in interrupt handler", exc_info=True)
            raise
        except Exception as ex:
            logger.error("Unhandled interpretation error in interrupt handler", exc_info=True)
            raise InterpretationError("Interpreter error") from ex

    def stop(self):
        self.running = False
        self._program.reset_runtime_state(recursive=True)

    def _is_awaiting_threshold(self, node: PNode):
        if isinstance(node, PInstruction) and node.time is not None and not node.forced:
            base_unit = self.context.tags.get(SystemTagName.BASE).get_value()
            assert isinstance(base_unit, str), \
                f"Base tag value must contain the base unit as a string. But its current value is '{base_unit}'"
            unit_provider = self.context.base_unit_provider
            if not unit_provider.has(base_unit):
                raise NodeInterpretationError(node, f"Base unit error. The current base unit '{base_unit}' is \
                                              not registered in the {term_uod}")

            value_tag, block_value_tag = unit_provider.get_tags(base_unit)
            if not self.context.tags.has(value_tag):
                raise InterpretationInternalError(f"Threshold calculation error. The registered tag \
                                              '{value_tag}' for unit '{base_unit}' is currently unavailable")
            if not self.context.tags.has(block_value_tag):
                raise InterpretationInternalError(f"Threshold calculation error. The registered block tag \
                                              '{block_value_tag}' for unit '{base_unit}' is currently unavailable")

            value_tag, block_value_tag = self.context.tags[value_tag], self.context.tags[block_value_tag]
            is_in_block = self.context.tags[SystemTagName.BLOCK].get_value() not in [None, ""]

            threshold_value = str(node.time)
            threshold_unit = base_unit

            time_value = str(block_value_tag.get_value() if is_in_block else value_tag.get_value())
            time_unit = block_value_tag.unit if is_in_block else value_tag.unit

            try:
                # calculate result of 'value < threshold'
                result = units.compare_values(
                    '<', time_value, time_unit,
                    threshold_value, threshold_unit)
                if result:
                    logger.debug(
                        f"Node {node} is awaiting threshold: {threshold_value}, " +
                        f"current: {time_value}, base unit: '{base_unit}'")
                    return True

                logger.debug(
                    f"Node {node} is done awaiting threshold {threshold_value}, " +
                    f"current: {time_value}, base unit: '{base_unit}'")
            except Exception as ex:
                raise NodeInterpretationError(
                    node,
                    "Threshold comparison error. Failed to compare " +
                    f"value '{time_value}' to threshold '{threshold_value}'") from ex
        return False

    def _evaluate_condition(self, condition_node: PWatch | PAlarm) -> bool:
        c = condition_node.condition
        assert c is not None, "Error in condition"
        assert not c.error, f"Error parsing condition '{c.condition_str}'"
        assert c.tag_name, "Error in condition tag"
        assert c.tag_value, "Error in condition value"
        assert self.context.tags.has(c.tag_name), f"Unknown tag '{c.tag_name}' in condition '{c.condition_str}'"
        tag = self.context.tags.get(c.tag_name)
        tag_value, tag_unit = str(tag.get_value()), tag.unit
        # TODO: Possible enhancement: if no unit specified, pick base unit?
        expected_value, expected_unit = c.tag_value, c.tag_unit

        return units.compare_values(
            c.op,
            tag_value,
            tag_unit,
            expected_value,
            expected_unit)

    def _visit_children(self, children: list[PNode] | None):
        ar = self.stack.peek()
        if children is not None:
            for child in children:
                if ar.complete:
                    break
                yield from self.visit(child)

    # Visitor Impl

    @override
    def visit(self, node: PNode):
        if not self.running:
            return

        # interrupts are visited multiple times. make sure to only create
        # a new record on the first visit
        record = self.runtimeinfo.get_last_node_record_or_none(node)
        if record is None:
            record = self.runtimeinfo.begin_visit(
                node,
                self._tick_time, self._tick_number,
                self.context.tags.as_readonly())

        # log all visits
        logger.debug(f"Visiting {node} at tick {self._tick_time}")

        # possibly wait for node threshold to expire
        while self._is_awaiting_threshold(node):
            if self.stack.peek().complete:
                logger.debug("Aborting threshold because parent block was complete")
                return

            record.add_state_awaiting_threshold(
                self._tick_time, self._tick_number,
                self.context.tags.as_readonly())
            yield

        # delegate to concrete visitor via base method
        result = super().visit(node)

        # allow visit methods to be non generators
        if inspect.isgenerator(result):
            yield from result
        else:
            yield

        record.set_end_visit(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())

        logger.debug(f"Visit {node} done")

    def visit_PProgram(self, node: PProgram):
        ar = ActivationRecord(node, self._tick_time)
        self.stack.push(ar)
        yield from self._visit_children(node.children)
        # TODO consider awaiting uod command completion before emit_on_method_end
        self.context.emitter.emit_on_method_end()
        self.stack.pop()

    def visit_PBlank(self, node: PBlank):
        pass

    def visit_PMark(self, node: PMark):
        record = self.runtimeinfo.get_last_node_record(node)

        logger.info(f"Mark {str(node)}")

        try:
            mark_tag = self.context.tags.get("Mark")
            mark_tag.set_value(node.name, self._tick_time)
        except ValueError:
            logger.error("Failed to get Mark tag")

        record.add_state_started(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())

        # we want this in but it breaks a lot of tests. we want to be able to manage that first
        # yield

        record.add_state_completed(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())

    def visit_PBlock(self, node: PBlock):
        record = self.runtimeinfo.get_last_node_record(node)

        ar = ActivationRecord(node, self._tick_time)
        self.stack.push(ar)
        self.context.tags[SystemTagName.BLOCK].set_value(node.name, self._tick_number)
        self.context.emitter.emit_on_block_start(node.name, self._tick_number)
        logger.debug(f"Block Tag set to {node.name}")

        yield  # comment if we dont want block to always consume a tick

        record.add_state_started(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())

        yield from self._visit_children(node.children)

        if not ar.complete:
            logger.debug(f"Block {node.name} idle")
        while not ar.complete and self.running:
            yield

        # await possible interrupt using the stack
        while not self.stack.peek() is ar:
            yield

        # now it's safe to pop
        self.stack.pop()

        record.add_state_completed(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())

        # restore previous block name
        block_restored = False
        for ar in self.stack.iterate_from_top():
            if ar.artype == ARType.BLOCK:
                assert isinstance(ar.owner, PBlock)
                self.context.tags[SystemTagName.BLOCK].set_value(ar.owner.name, self._tick_number)
                self.context.emitter.emit_on_block_end(node.name, ar.owner.name, self._tick_number)
                logger.debug(f"Block Tag popped to {ar.owner.name}")
                block_restored = True
        if not block_restored:
            self.context.tags[SystemTagName.BLOCK].set_value(None, self._tick_number)
            self.context.emitter.emit_on_block_end(node.name, "", self._tick_number)
            logger.debug(f"Block Tag cleared from {node.name}")


    def visit_PEndBlock(self, node: PEndBlock):
        record = self.runtimeinfo.get_last_node_record(node)
        record.add_state_started(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())
        for ar in reversed(self.stack.records):
            if ar.artype == ARType.BLOCK:
                ar.complete = True
                record.add_state_completed(
                    self._tick_time, self._tick_number,
                    self.context.tags.as_readonly())
                return
        logger.warning("End block found no block to end")

    def visit_PEndBlocks(self, node: PEndBlocks):
        record = self.runtimeinfo.get_last_node_record(node)
        record.add_state_started(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())
        for ar in reversed(self.stack.records):
            if ar.artype == ARType.BLOCK:
                ar.complete = True
                logger.debug(f"EndBlocks ended block {ar.owner.display_name}")
        record.add_state_completed(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())

    def execute_interpreter_command(self, node: PCommand):
        record = self.runtimeinfo.get_last_node_record(node)
        record.add_state_started(self._tick_time, self._tick_number, self.context.tags.as_readonly())

        if node.name == InterpreterCommandEnum.BASE:
            valid_units = self.context.base_unit_provider.get_units()
            if node.args is None or node.args not in valid_units:
                record.add_state_failed(self._tick_time, self._tick_number, self.context.tags.as_readonly())
                raise NodeInterpretationError(node, f"Base instruction has invalid argument '{node.args}'. \
                    Value must be one of {', '.join(valid_units)}")
            self.context.tags[SystemTagName.BASE].set_value(node.args, self._tick_time)

        elif node.name == InterpreterCommandEnum.INCREMENT_RUN_COUNTER:
            run_counter = self.context.tags[SystemTagName.RUN_COUNTER]
            rc_value = run_counter.as_number() + 1
            logger.debug(f"Run Counter incremented from {rc_value - 1} to {rc_value}")
            run_counter.set_value(rc_value, self._tick_time)

        elif node.name == InterpreterCommandEnum.RUN_COUNTER:
            try:
                new_value = int(node.args)
                logger.debug(f"Run Counter set to {new_value}")
                self.context.tags[SystemTagName.RUN_COUNTER].set_value(new_value, self._tick_time)
            except ValueError:
                raise NodeInterpretationError(node, f"Invalid argument '{node.args}'. Argument must be an integer")

        else:
            record.add_state_failed(self._tick_time, self._tick_number, self.context.tags.as_readonly())
            raise NodeInterpretationError(node, f"Interpreter command '{node.name}' is not supported")

        record.add_state_completed(self._tick_time, self._tick_number, self.context.tags.as_readonly())

    def visit_PCommand(self, node: PCommand):
        if InterpreterCommandEnum.has_value(node.name):
            logger.debug(f"Executing interpreter command {str(node)}")
            self.execute_interpreter_command(node)
        else:
            record = self.runtimeinfo.get_last_node_record(node)

            # Note: Commands can be resident and last multiple ticks.
            # The context (Engine) keeps track of this and we just
            # move on to the next instruction when tick() is invoked.
            # We do, however, provide the execution id to the context
            # so that it can update the runtime record appropriately.
            try:
                logger.debug(f"Executing command '{str(node)}' with args '{node.args}' via engine")
                self.context.schedule_execution(
                    name=node.name,
                    exec_id=record.exec_id,
                    unparsed_args=node.args)
            except Exception as ex:
                record.add_state_failed(
                    self._tick_time, self._tick_number,
                    self.context.tags.as_readonly())

                if isinstance(ex, EngineError):
                    raise
                else:
                    raise NodeInterpretationError(node, "Failed to pass command to engine") from ex

        yield

    def visit_PCommandWithDuration(self, node: PCommandWithDuration):
        record = self.runtimeinfo.get_last_node_record(node)

        if node.duration is not None and node.duration.error:
            record.add_state_failed(self._tick_time, self._tick_number, self.context.tags.as_readonly())
            raise NodeInterpretationError(node, "Parse error. Command duration not valid") from None
        try:
            logger.debug(f"Executing command '{str(node)}' via engine")
            if node.duration is None:
                self.context.schedule_execution(name=node.name, exec_id=record.exec_id)
            else:
                self.context.schedule_execution(
                    name=node.name,
                    exec_id=record.exec_id,
                    time=node.duration.time,
                    unit=node.duration.unit
                )
        except Exception as ex:
            record.add_state_failed(
                self._tick_time, self._tick_number,
                self.context.tags.as_readonly())
            raise NodeInterpretationError(node, "Failed to pass command to engine") from ex

        yield


    def visit_PWatch(self, node: PWatch):
        yield from self.visit_WatchOrAlarm(node)

    def visit_PAlarm(self, node: PAlarm):
        yield from self.visit_WatchOrAlarm(node)

    def visit_WatchOrAlarm(self, node: PWatch | PAlarm):
        record = self.runtimeinfo.get_last_node_record(node)

        ar = self.stack.peek()
        if ar.owner is not node:
            ar = ActivationRecord(node)
            self._register_interrupt(ar, self._create_interrupt_handler(node, ar))
            record.add_state_awaiting_interrupt(
                self._tick_time, self._tick_number,
                self.context.tags.as_readonly())
        else:
            # On subsequent visits (originating from the interrupt handler), we evaluate
            # the condition. When true, the node is 'activated' and its body will run
            logger.debug(f"{str(node)} interrupt invoked")
            if node.activated:
                logger.debug(f"{str(node)} was previously activated")
            else:
                record.add_state_awaiting_condition(
                    self._tick_time, self._tick_number,
                    self.context.tags.as_readonly())

                while not ar.complete and self.running:
                    condition_result = False
                    if node.cancelled:
                        record.add_state_cancelled(self._tick_time, self._tick_number, self.context.tags.as_readonly())
                        logger.info(f"Instruction {node} cancelled")
                        ar.complete = True
                        return
                    elif node.forced:
                        record.add_state_forced(self._tick_time, self._tick_number, self.context.tags.as_readonly())
                        logger.info(f"Instruction {node} forced")
                        condition_result = True
                    else:
                        try:
                            condition_result = self._evaluate_condition(node)
                        except AssertionError:
                            raise
                        except Exception as ex:
                            raise NodeInterpretationError(node, "Error evaluating condition: " + str(ex))
                        logger.debug(f"{str(node)} condition evaluated: {condition_result}")

                    if condition_result:
                        node.activated = True
                        logger.debug(f"{str(node)} executing")
                        record.add_state_started(
                            self._tick_time, self._tick_number,
                            self.context.tags.as_readonly())

                        yield from self._visit_children(node.children)
                        logger.debug(f"{str(node)} executed")
                        ar.complete = True
                        record.add_state_completed(
                            self._tick_time, self._tick_number,
                            self.context.tags.as_readonly())

                        logger.info(f"Interrupt complete {ar}")
                    else:
                        yield

    def visit_PInjectedNode(self, node: PInjectedNode):
        record = self.runtimeinfo.get_last_node_record(node)

        ar = self.stack.peek()
        if ar.owner is node:
            logger.debug(f"{str(node)} injected interrupt invoked")
            record.add_state_started(
                self._tick_time, self._tick_number,
                self.context.tags.as_readonly()
            )
            while not ar.complete and self.running:
                logger.debug(f"{str(node)} executing")
                yield from self._visit_children(node.children)
                logger.debug(f"{str(node)} executed")
                ar.complete = True
                logger.info(f"Injected interrupt complete {ar}")
                record.add_state_completed(
                    self._tick_time, self._tick_number,
                    self.context.tags.as_readonly())
        else:
            logger.error("Injected node error. Stack unwind required?")
            record.add_state_failed(
                self._tick_time, self._tick_number,
                self.context.tags.as_readonly())

    def visit_PComment(self, node: PComment):
        pass

    def visit_PErrorInstruction(self, node: PErrorInstruction):
        record = self.runtimeinfo.get_last_node_record(node)

        logger.error(f"Invalid instruction: {str(node)}")

        record.add_state_started(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())
        record.add_state_failed(
            self._tick_time, self._tick_number,
            self.context.tags.as_readonly())

        raise NodeInterpretationError(node, f"Invalid instruction '{node.code}'")



import logging
from typing import Protocol
from uuid import UUID

import openpectus.sentry as sentry
import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.messages as M
from openpectus.engine.engine import Engine
from openpectus.protocol.engine_dispatcher import EngineMessageHandler


logger = logging.getLogger(__name__)


class EngineDispatcherRpcHandler(Protocol):
    def set_rpc_handler(self, message_type: type[AM.AggregatorMessage], handler: EngineMessageHandler):
        ...


class EngineMessageHandlers():
    def __init__(self, engine: Engine, dispatcher: EngineDispatcherRpcHandler) -> None:
        self.engine = engine
        dispatcher.set_rpc_handler(AM.InvokeCommandMsg, self.handle_invokeCommandMsg)
        dispatcher.set_rpc_handler(AM.InjectCodeMsg, self.handle_injectCodeMsg)
        dispatcher.set_rpc_handler(AM.MethodMsg, self.handle_methodMsg)
        dispatcher.set_rpc_handler(AM.CancelMsg, self.handle_cancelMsg)
        dispatcher.set_rpc_handler(AM.ForceMsg, self.handle_forceMsg)

    async def handle_methodMsg(self, msg: AM.AggregatorMessage):
        assert isinstance(msg, AM.MethodMsg)
        logger.info("Incomming set_method command from aggregator")
        if self.engine._runstate_started:
            logger.error("Cannot modify running method - yet")
            return AM.ErrorMessage(message="Cannot modify running method")
        try:
            self.engine.set_method(msg.method)
            sentry.engine_method_set(msg.method.as_pcode())
            return AM.SuccessMessage()
        except Exception as ex:
            logger.error("Failed to set method")
            return AM.ErrorMessage(message="Failed to set method", exception_message=str(ex))

    async def handle_invokeCommandMsg(self, msg: AM.AggregatorMessage) -> M.MessageBase:
        assert isinstance(msg, AM.InvokeCommandMsg)
        logger.info(f"Incomming command from aggregator: {msg.name}")
        try:
            self.engine.schedule_execution_user(name=msg.name, args=msg.arguments)
            return AM.SuccessMessage()
        except Exception:
            logger.error(f"The command '{msg.name}' could not be scheduled")
            return AM.ErrorMessage(message=f"The command '{msg.name}' could not be scheduled")

    async def handle_injectCodeMsg(self, msg: AM.AggregatorMessage) -> M.MessageBase:
        assert isinstance(msg, AM.InjectCodeMsg)
        logger.info(f"Incomming inject code command from aggregator: '{msg.pcode}'")
        try:
            self.engine.inject_code(msg.pcode)
            return AM.SuccessMessage()
        except Exception:
            logger.error("Code injection failed")
            return AM.ErrorMessage(message="The code could not be injected")

    async def handle_cancelMsg(self, msg: AM.AggregatorMessage) -> M.MessageBase:
        assert isinstance(msg, AM.CancelMsg)
        logger.info(f"Incomming cancel request {msg.exec_id}")
        try:
            exec_id = UUID(msg.exec_id)
            self.engine.cancel_instruction(exec_id)
            return AM.SuccessMessage()
        except Exception:
            logger.error("Cancel failed", exc_info=True)
            return AM.ErrorMessage(message="The engine failed to cancel the instruction")

    async def handle_forceMsg(self, msg: AM.AggregatorMessage) -> M.MessageBase:
        assert isinstance(msg, AM.ForceMsg)
        logger.info(f"Incomming force request {msg.exec_id}")
        try:
            exec_id = UUID(msg.exec_id)
            self.engine.force_instruction(exec_id)
            return AM.SuccessMessage()
        except Exception:
            logger.error("Force failed", exc_info=True)
            return AM.ErrorMessage(message="The engine failed to force the instruction")



"""sad contributors to recent engine

Revision ID: cfb648a71b50
Revises: bcda1e0dd461
Create Date: 2024-12-16 13:58:44.694728

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'cfb648a71b50'
down_revision: Union[str, None] = 'bcda1e0dd461'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentEngines', schema=None) as batch_op:
        batch_op.add_column(sa.Column('contributors', sa.JSON(), nullable=False, server_default='[]'))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentEngines', schema=None) as batch_op:
        batch_op.drop_column('contributors')

    # ### end Alembic commands ###
